import os
from datetime import datetime, timedelta
from pathlib import Path
from feathr import FeathrClient
from feathr import (BackfillTime, MaterializationSettings)
from feathr import FeathrClient

from feathr import HdfsSink
from feathr.utils.job_utils import get_result_df, copy_cloud_dir, cloud_dir_exists
from test_fixture import (basic_test_setup, time_partition_pattern_feature_gen_test_setup, time_partition_pattern_feature_join_test_setup)
from test_utils.constants import Constants
'''
def setup_module():
    """
    Prepare data sources for 'timePartitionPattern' test cases
    """
    test_workspace_dir = Path(
        __file__).parent.resolve() / "test_user_workspace"
   
    # Create data sources to support testing with 'timePartitionPattern' cases below
    client_producer: FeathrClient = basic_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))

    if client_producer.spark_runtime == 'databricks':
        output_path = 'dbfs:/timePartitionPattern_test'
        output_pf_path = 'dbfs:/timePartitionPattern_postfix_test/df0/daily/2020/05/01/postfixPath'
        output_hourly_path = 'dbfs:/timePartitionPattern_hourly_test/df0/daily/2020/05/01/00'
    else:
        output_path = 'abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/timePartitionPattern_test'
        output_pf_path = 'abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/timePartitionPattern_postfix_test/df0/daily/2020/05/01/postfixPath'
        output_hourly_path = 'abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/timePartitionPattern_hourly_test/df0/daily/2020/05/01/00'
     
    source_url = output_path + "/df0/daily/2020/05/01" 
    if not cloud_dir_exists(client_producer, source_url):
        backfill_time = BackfillTime(start=datetime(
            2020, 5, 1), end=datetime(2020, 5, 1), step=timedelta(days=1))
        offline_sink = HdfsSink(output_path=output_path)
        settings = MaterializationSettings("nycTaxiTable",
                sinks=[offline_sink],
                feature_names=[
                    "f_location_avg_fare", "f_location_max_fare"],
                backfill_time=backfill_time)
    
        client_producer.materialize_features(settings)
        # assuming the job can successfully run; otherwise it will throw exception
        client_producer.wait_job_to_finish(timeout_sec=Constants.SPARK_JOB_TIMEOUT_SECONDS)
    
        # Check if data sources prepared well
        res_df = get_result_df(client_producer, data_format="avro", res_url=source_url)
        assert res_df.shape[0] > 0
    
    # Copy created data sources to another folder to support 'postfix_path' test  
    if not cloud_dir_exists(client_producer, output_pf_path):
        copy_cloud_dir(client_producer, source_url, output_pf_path)
        res_df_pf = get_result_df(client_producer, data_format="avro", res_url=output_pf_path)
        assert res_df_pf.shape[0] > 0
    
    # Copy created data sources to another folder to support 'hourly' test
    if not cloud_dir_exists(client_producer, output_hourly_path):
        copy_cloud_dir(client_producer, source_url, output_hourly_path)
        res_df_hourly = get_result_df(client_producer, data_format="avro", res_url=output_hourly_path)
        assert res_df_hourly.shape[0] > 0
'''
def test_feathr_materialize_with_time_partition_pattern():
    """
    Test FeathrClient() using HdfsSource with 'timePartitionPattern'.
    """
    test_workspace_dir = Path(
        __file__).parent.resolve() / "test_user_workspace"
    
    client_dummy = FeathrClient(os.path.join(test_workspace_dir, "feathr_config.yaml"))
    if client_dummy.spark_runtime == 'databricks':
        source_path = 'dbfs:/timePartitionPattern_test/df0/daily/'
    else:
        source_path = 'abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/timePartitionPattern_test/df0/daily/'
      
    client: FeathrClient = time_partition_pattern_feature_gen_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"), source_path,local_workspace_dir="test_materialize_tpp")
    
    backfill_time_tpp = BackfillTime(start=datetime(
        2020, 5, 2), end=datetime(2020, 5, 2), step=timedelta(days=1))
    now = datetime.now()
    if client.spark_runtime == 'databricks':
        output_path_tpp = ''.join(['dbfs:/feathrazure_cijob_materialize_offline_','_', str(now.minute), '_', str(now.second), ""])
    else:
        output_path_tpp = ''.join(['abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/demo_data/feathrazure_cijob_materialize_offline_','_', str(now.minute), '_', str(now.second), ""])
    offline_sink_tpp = HdfsSink(output_path=output_path_tpp)
    settings_tpp = MaterializationSettings("nycTaxiTable",
                                       sinks=[offline_sink_tpp],
                                       feature_names=[
                                           "f_loc_avg_output", "f_loc_max_output"],
                                       backfill_time=backfill_time_tpp)
    client.materialize_features(settings_tpp)
    client.wait_job_to_finish(timeout_sec=Constants.SPARK_JOB_TIMEOUT_SECONDS)
    
    res_df = get_result_df(client, data_format="avro", res_url=output_path_tpp + "/df0/daily/2020/05/02")
    assert res_df.shape[0] > 0
     
def test_feathr_materialize_with_time_partition_pattern_postfix_path():
    """
    Test FeathrClient() using HdfsSource with 'timePartitionPattern' and 'postfixPath'.
    """  
    test_workspace_dir = Path(
        __file__).parent.resolve() / "test_user_workspace"
   
    client_dummy = FeathrClient(os.path.join(test_workspace_dir, "feathr_config.yaml"))
    if client_dummy.spark_runtime == 'databricks':
        source_path = 'dbfs:/timePartitionPattern_postfix_test/df0/daily/'
    else:
        source_path = 'abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/timePartitionPattern_postfix_test/df0/daily/'
    
    client: FeathrClient = time_partition_pattern_feature_gen_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"), source_path,local_workspace_dir="test_materialize_tpp_postfix", postfix_path='postfixPath')

    backfill_time_pf = BackfillTime(start=datetime(
        2020, 5, 2), end=datetime(2020, 5, 2), step=timedelta(days=1))
    now = datetime.now()
    if client.spark_runtime == 'databricks':
        output_path_pf = ''.join(['dbfs:/feathrazure_cijob_materialize_offline_','_', str(now.minute), '_', str(now.second), ""])
    else:
        output_path_pf = ''.join(['abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/demo_data/feathrazure_cijob_materialize_offline_','_', str(now.minute), '_', str(now.second), ""])
    offline_sink_pf = HdfsSink(output_path=output_path_pf)
    settings_pf = MaterializationSettings("nycTaxiTable",
                                       sinks=[offline_sink_pf],
                                       feature_names=[
                                           "f_loc_avg_output", "f_loc_max_output"],
                                       backfill_time=backfill_time_pf)
    client.materialize_features(settings_pf)
    client.wait_job_to_finish(timeout_sec=Constants.SPARK_JOB_TIMEOUT_SECONDS)

    res_df = get_result_df(client, data_format="avro", res_url=output_path_pf + "/df0/daily/2020/05/02")
    assert res_df.shape[0] > 0
    
def test_feathr_materialize_with_time_partition_pattern_hourly():
    """
    Test FeathrClient() using HdfsSource with hourly 'timePartitionPattern'.
    """   
    test_workspace_dir = Path(
        __file__).parent.resolve() / "test_user_workspace"
   
    client_dummy = FeathrClient(os.path.join(test_workspace_dir, "feathr_config.yaml"))
    if client_dummy.spark_runtime == 'databricks':
        source_path = 'dbfs:/timePartitionPattern_hourly_test/df0/daily/'
    else:
        source_path = 'abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/timePartitionPattern_hourly_test/df0/daily/'
    
    client: FeathrClient = time_partition_pattern_feature_gen_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"), source_path,local_workspace_dir="test_materialize_hourly", resolution='HOURLY')

    backfill_time_tpp = BackfillTime(start=datetime(
        2020, 5, 2), end=datetime(2020, 5, 2), step=timedelta(days=1))
    now = datetime.now()
    if client.spark_runtime == 'databricks':
        output_path_tpp = ''.join(['dbfs:/feathrazure_cijob_materialize_offline_','_', str(now.minute), '_', str(now.second), ""])
    else:
        output_path_tpp = ''.join(['abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/demo_data/feathrazure_cijob_materialize_offline_','_', str(now.minute), '_', str(now.second), ""])
    offline_sink_tpp = HdfsSink(output_path=output_path_tpp)
    settings_tpp = MaterializationSettings("nycTaxiTable",
                                       sinks=[offline_sink_tpp],
                                       feature_names=[
                                           "f_loc_avg_output", "f_loc_max_output"],
                                       backfill_time=backfill_time_tpp,
                                       resolution = 'HOURLY')
    client.materialize_features(settings_tpp)
    client.wait_job_to_finish(timeout_sec=Constants.SPARK_JOB_TIMEOUT_SECONDS)

    res_df = get_result_df(client, data_format="avro", res_url=output_path_tpp + "/df0/daily/2020/05/02/00")
    assert res_df.shape[0] > 0

def test_feathr_get_offline_with_time_partition_pattern_postfix_path():
    """
    Test FeathrClient() using HdfsSource with 'timePartitionPattern' and 'postfixPath'.
    """  
    test_workspace_dir = Path(
        __file__).parent.resolve() / "test_user_workspace"
   
    client_dummy = FeathrClient(os.path.join(test_workspace_dir, "feathr_config.yaml"))
    if client_dummy.spark_runtime == 'databricks':
        source_path = 'dbfs:/timePartitionPattern_postfix_test/df0/daily/'
    else:
        source_path = 'abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/timePartitionPattern_postfix_test/df0/daily/'
    
    [client, feature_query, settings] = time_partition_pattern_feature_join_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"), source_path, local_workspace_dir="test_offline_tpp",postfix_path='postfixPath')

    now = datetime.now()
    if client.spark_runtime == 'databricks':
        output_path = ''.join(['dbfs:/feathrazure_cijob','_', str(now.minute), '_', str(now.second), ".avro"])
    else:
        output_path = ''.join(['abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/demo_data/output','_', str(now.minute), '_', str(now.second), ".avro"])
                
    client.get_offline_features(observation_settings=settings,
                            feature_query=feature_query,
                            output_path=output_path)
    client.wait_job_to_finish(timeout_sec=Constants.SPARK_JOB_TIMEOUT_SECONDS)

    res_df = get_result_df(client, data_format="avro", res_url = output_path)
    assert res_df.shape[0] > 0
    



import os
from pathlib import Path
import pytest
from feathr import MaterializationSettings, RedisSink, AvroJsonSchema, KafkaConfig, KafKaSource
from test_fixture import kafka_test_setup
from test_utils.constants import Constants

def test_feathr_kafka_configs():
    schema = AvroJsonSchema(schemaStr="""
    {
        "type": "record",
        "schema_name": "DriverTrips",
        "fields": [
            {"name": "driver_id", "type": "long"},
        ]
    }
    """)
    stream_source = KafKaSource(name="kafkaStreamingSource",
                              kafkaConfig=KafkaConfig(brokers=["feathrazureci.servicebus.windows.net:9093"],
                                                      topics=["feathrcieventhub"],
                                                      schema=schema))
    config_list = stream_source.to_feature_config().split()
    config = ''.join([conf.replace('\\"', '').strip('\\n') for conf in config_list])
    expected_config = """
    kafkaStreamingSource: {
    type: KAFKA
    config: {
        brokers: ["feathrazureci.servicebus.windows.net:9093"]
        topics: [feathrcieventhub]
        
        schema: {
            type = "avro"
            avroJson:"{
                type: record,
                schema_name: DriverTrips,
                fields: [{
                    name: driver_id, 
                    type: long
                    },]
            }"
        }  
    }
    }
    """
    assert config == "".join(expected_config.split())
    
@pytest.mark.skipif(os.environ.get('SPARK_CONFIG__SPARK_CLUSTER') != "azure_synapse",
                    reason="skip for databricks, as it cannot stop streaming job automatically for now.")
def test_feathr_kafa_streaming_features():
    """
    Test FeathrClient() materialize_features can ingest streaming feature correctly
    """
    test_workspace_dir = Path(__file__).parent.resolve() / "test_user_workspace"

    client = kafka_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))
    redisSink = RedisSink(table_name="kafkaSampleDemoFeature", streaming=True, streamingTimeoutMs=10000)
    settings = MaterializationSettings(name="kafkaSampleDemo",
                                   sinks=[redisSink],
                                   feature_names=['f_modified_streaming_count']
                                   )
    client.materialize_features(settings, allow_materialize_non_agg_feature=True)
    client.wait_job_to_finish(timeout_sec=Constants.SPARK_JOB_TIMEOUT_SECONDS)



from logging import PlaceHolder
from typing import List
from feathr.definition.sink import Sink
from feathr.definition.source import Source

class FeatureJoinJobParams:
    """Parameters related to feature join job.

    Attributes:
        join_config_path: Path to the join config.
        observation_path: Absolute path in Cloud to the observation data path.
        feature_config: Path to the features config.
        job_output_path: Absolute path in Cloud that you want your output data to be in.
    """

    def __init__(self, join_config_path, observation_path, feature_config, job_output_path, secrets:List[str]=[]):
        self.secrets = secrets
        self.join_config_path = join_config_path
        if isinstance(observation_path, str):
            self.observation_path = observation_path
        elif isinstance(observation_path, Source):
            self.observation_path = observation_path.to_argument()
            if hasattr(observation_path, "get_required_properties"):
                self.secrets.extend(observation_path.get_required_properties())
        else:
            raise TypeError("observation_path must be a string or a Sink")
        self.feature_config = feature_config
        if isinstance(job_output_path, str):
            self.job_output_path = job_output_path
        elif isinstance(job_output_path, Sink):
            self.job_output_path = job_output_path.to_argument()
            if hasattr(job_output_path, "get_required_properties"):
                self.secrets.extend(job_output_path.get_required_properties())
        else:
            raise TypeError("job_output_path must be a string or a Sink")

class FeatureGenerationJobParams:
    """Parameters related to feature generation job.

    Attributes:
        generation_config_path: Path to the feature generation config.
        feature_config: Path to the features config.
        secrets: secret names from data sources, the values will be taken from env or KeyVault
    """

    def __init__(self, generation_config_path, feature_config, secrets=[]):
        self.generation_config_path = generation_config_path
        self.feature_config = feature_config
        self.secrets = secrets



import os


def is_jupyter() -> bool:
    """Check if the module is running on Jupyter notebook/console.
    Note - there might be better way to check if the code is running on a jupyter notebook or not,
    but this hacky way still works.

    Ref:
        https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook

    Returns:
        bool: True if the module is running on Jupyter notebook or Jupyter console, False otherwise.
    """
    # Should check is_databricks() and is_synapse() first since they also use the same ZMQ interactive shell.
    if is_databricks() or is_synapse():
        return False

    try:
        # Pre-loaded module `get_ipython()` tells you whether you are running inside IPython or not.
        shell_name = get_ipython().__class__.__name__
        # `ZMQInteractiveShell` tells you if this is an interactive mode (notebook).
        if shell_name == "ZMQInteractiveShell":
            return True
        else:
            return False
    except NameError:
        return False


def is_databricks() -> bool:
    """Check if the module is running on Databricks.

    Returns:
        bool: True if the module is running on Databricks notebook, False otherwise.
    """
    # Note, this is a hacky way to check if the code is running on Databricks.
    if "DATABRICKS_RUNTIME_VERSION" in os.environ:
        return True
    else:
        return False


def is_synapse() -> bool:
    """Check if the module is running on Azure Synapse.

    Returns:
        bool: True if the module is running on Azure Synapse notebook, False otherwise.
    """
    # Note, this is a hacky way to check if the code is running on Synapse.
    if "SYNAPSE_ENABLE_CONFIG_MERGE_RULE" in os.environ:
        return True
    else:
        return False



import os
import time
from datetime import datetime
from pathlib import Path

from feathr import (FeatureQuery, ObservationSettings, SparkExecutionConfiguration, TypedKey, ValueType)
from feathr.client import FeathrClient
from feathr.constants import OUTPUT_FORMAT
from feathr.utils.job_utils import get_result_df
from test_fixture import basic_test_setup
from test_utils.constants import Constants


# test parquet file read/write without an extension name
def test_feathr_get_offline_features_with_parquet():
    """
    Test if the program can read and write parquet files
    """

    test_workspace_dir = Path(
        __file__).parent.resolve() / "test_user_workspace"

    client: FeathrClient = basic_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))

    location_id = TypedKey(key_column="DOLocationID",
                            key_column_type=ValueType.INT32)

    feature_query = FeatureQuery(
        feature_list=["f_location_avg_fare"], key=location_id)
    settings = ObservationSettings(
        observation_path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04",
        event_timestamp_column="lpep_dropoff_datetime",
        timestamp_format="yyyy-MM-dd HH:mm:ss")

    now = datetime.now()
    # set output folder based on different runtime
    if client.spark_runtime == 'databricks':
        output_path = ''.join(['dbfs:/feathrazure_cijob','_', str(now.minute), '_', str(now.second),'_', str(now.microsecond),  ".parquet"])
    else:
        output_path = ''.join(['abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/demo_data/output','_', str(now.minute), '_', str(now.second), ".parquet"])


    client.get_offline_features(observation_settings=settings,
                                feature_query=feature_query,
                                output_path=output_path,
                                execution_configurations=SparkExecutionConfiguration({"spark.feathr.inputFormat": "parquet", "spark.feathr.outputFormat": "parquet"})
                                )

    # assuming the job can successfully run; otherwise it will throw exception
    client.wait_job_to_finish(timeout_sec=Constants.SPARK_JOB_TIMEOUT_SECONDS)

    # download result and just assert the returned result is not empty
    res_df = get_result_df(client)
    assert res_df.shape[0] > 0


# test delta lake read/write without an extension name
def test_feathr_get_offline_features_with_delta_lake():
    """
    Test if the program can read and write delta lake
    """

    test_workspace_dir = Path(
        __file__).parent.resolve() / "test_user_workspace"

    client = basic_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))

    location_id = TypedKey(key_column="DOLocationID",
                            key_column_type=ValueType.INT32)

    feature_query = FeatureQuery(
        feature_list=["f_location_avg_fare"], key=location_id)
    settings = ObservationSettings(
        observation_path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/feathr_delta_table",
        event_timestamp_column="lpep_dropoff_datetime",
        timestamp_format="yyyy-MM-dd HH:mm:ss")

    now = datetime.now()
    # set output folder based on different runtime
    if client.spark_runtime == 'databricks':
        output_path = ''.join(['dbfs:/feathrazure_cijob','_', str(now.minute), '_', str(now.second), "_deltalake"])
    else:
        output_path = ''.join(['abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/demo_data/output','_', str(now.minute), '_', str(now.second), "_deltalake"])


    client.get_offline_features(observation_settings=settings,
                                feature_query=feature_query,
                                output_path=output_path,
                                execution_configurations=SparkExecutionConfiguration({"spark.feathr.inputFormat": "delta", "spark.feathr.outputFormat": "delta"})
                                )

    # assuming the job can successfully run; otherwise it will throw exception
    client.wait_job_to_finish(timeout_sec=Constants.SPARK_JOB_TIMEOUT_SECONDS)

    # wait for a few secs for the resource to come up in the databricks API
    time.sleep(5)

    # download result and just assert the returned result is not empty
    # if users are using delta format in synapse, skip this check, due to issue https://github.com/delta-io/delta-rs/issues/582
    result_format: str = client.get_job_tags().get(OUTPUT_FORMAT, "")
    if not (client.spark_runtime == 'azure_synapse' and result_format == 'delta'):
        res_df = get_result_df(client)
        assert res_df.shape[0] > 0



import collections.abc
from copy import deepcopy
import os
import json
from tempfile import NamedTemporaryFile
from typing import Dict
import yaml

from feathr.utils.platform import is_databricks


DEFAULT_FEATHR_CONFIG = {
    "api_version": 1,
    "project_config": {},  # "project_name"
    "feature_registry": {},  # "api_endpoint"
    "spark_config": {
        "spark_cluster": "local",  # Currently support 'azure_synapse', 'databricks', and 'local'
        "spark_result_output_parts": "1",
    },
    "offline_store": {
        "adls": {"adls_enabled": "true"},
        "wasb": {"wasb_enabled": "true"},
    },
    "online_store": {
        "redis": {
            # "host"
            "port": "6380",
            "ssl_enabled": "true",
        }
    }
}

# New databricks job cluster config
DEFAULT_DATABRICKS_CLUSTER_CONFIG = {
    "spark_version": "11.2.x-scala2.12",
    "node_type_id": "Standard_DS3_v2",  # Change this if necessary
    "num_workers": 1,
    "spark_conf": {
        "FEATHR_FILL_IN": "FEATHR_FILL_IN",
        # Exclude conflicting packages if use feathr <= v0.8.0:
        "spark.jars.excludes": "commons-logging:commons-logging,org.slf4j:slf4j-api,com.google.protobuf:protobuf-java,javax.xml.bind:jaxb-api",
    },
}

# New Azure Synapse spark pool config
DEFAULT_AZURE_SYNAPSE_SPARK_POOL_CONFIG = {
    "executor_size": "Small",
    "executor_num": 1,
}


def generate_config(
    resource_prefix: str,
    project_name: str,
    output_filepath: str = None,
    databricks_workspace_token_value: str = None,
    databricks_cluster_id: str = None,
    redis_password: str = None,
    adls_key: str = None,
    **kwargs,
) -> str:
    """Generate a feathr config yaml file. Note, Feathr client will try to read environment variables first before read the config file.
    See details from the FeathrClient docstrings.

    Note:
        This utility function assumes Azure resources are deployed using the Azure Resource Manager (ARM) template,
        and infers resource names based on the given `resource_prefix`. If you deploy resources manually, you may need
        to pass each resource url manually, e.g. `spark_config__azure_synapse__dev_url="your-resource-url"`.

    Args:
        resource_prefix: Resource name prefix used when deploying Feathr resources by using ARM template.
        project_name: Feathr project name.
        output_filepath (optional): Output filepath.
        databricks_workspace_token_value (optional): Databricks workspace token. If provided, the value will be stored
            as the environment variable.
        databricks_cluster_id (optional): Databricks cluster id to use an existing cluster.
        redis_password (optional): Redis password. If provided, the value will be stored as the environment variable.
        adls_key (optional): ADLS key. If provided, the value will be stored as the environment variable.
        **kwargs: Keyword arguments to update the config. Keyword arguments follow the same naming convention as
            the feathr config. E.g. to set Databricks as the target cluster,
            use `spark_config__spark_cluster="databricks"`.
            See https://feathr-ai.github.io/feathr/quickstart_synapse.html#step-4-update-feathr-config for more details.

    Returns:
        str: Generated config file path. This will be identical to `output_filepath` if provided.
    """
    # Set keys
    if databricks_workspace_token_value:
        os.environ["DATABRICKS_WORKSPACE_TOKEN_VALUE"] = databricks_workspace_token_value
    if redis_password:
        os.environ["REDIS_PASSWORD"] = redis_password
    if adls_key:
        os.environ["ADLS_KEY"] = adls_key

    # Set configs
    config = deepcopy(DEFAULT_FEATHR_CONFIG)

    # Maybe update configs with environment variables
    _maybe_update_config_with_env_var(config, "SPARK_CONFIG__SPARK_CLUSTER")
    _maybe_update_config_with_env_var(config, "SPARK_CONFIG__AZURE_SYNAPSE__DEV_URL")
    _maybe_update_config_with_env_var(config, "SPARK_CONFIG__AZURE_SYNAPSE__POOL_NAME")
    _maybe_update_config_with_env_var(config, "SPARK_CONFIG__AZURE_SYNAPSE__WORKSPACE_DIR")
    _maybe_update_config_with_env_var(config, "SPARK_CONFIG__DATABRICKS__WORK_DIR")
    _maybe_update_config_with_env_var(config, "SPARK_CONFIG__DATABRICKS__WORKSPACE_INSTANCE_URL")
    _maybe_update_config_with_env_var(config, "SPARK_CONFIG__DATABRICKS__CONFIG_TEMPLATE")

    config["project_config"]["project_name"] = project_name
    config["feature_registry"]["api_endpoint"] = f"https://{resource_prefix}webapp.azurewebsites.net/api/v1"
    config["online_store"]["redis"]["host"] = f"{resource_prefix}redis.redis.cache.windows.net"

    # Update configs using kwargs
    new_config = _config_kwargs_to_dict(**kwargs)
    _update_config(config, new_config)

    # Set platform specific configurations
    if config["spark_config"]["spark_cluster"] == "local":
        _set_local_spark_config()
    elif config["spark_config"]["spark_cluster"] == "azure_synapse":
        _set_azure_synapse_config(
            config=config,
            resource_prefix=resource_prefix,
            project_name=project_name,
        )
    elif config["spark_config"]["spark_cluster"] == "databricks":
        _set_databricks_config(
            config=config,
            project_name=project_name,
            cluster_id=databricks_cluster_id,
        )

    # Verify config
    _verify_config(config)

    # Write config to file
    if not output_filepath:
        output_filepath = NamedTemporaryFile(mode="w", delete=False).name

    with open(output_filepath, "w") as f:
        yaml.dump(config, f, default_flow_style=False)

    return output_filepath


def _set_local_spark_config():
    """Set environment variables for local spark cluster."""
    os.environ["SPARK_LOCAL_IP"] = os.getenv(
        "SPARK_LOCAL_IP",
        "127.0.0.1",
    )


def _set_azure_synapse_config(
    config: Dict,
    resource_prefix: str,
    project_name: str,
):
    """Set configs for Azure Synapse spark cluster."""

    config["spark_config"]["azure_synapse"] = config["spark_config"].get("azure_synapse", {})

    if not config["spark_config"]["azure_synapse"].get("dev_url"):
        config["spark_config"]["azure_synapse"]["dev_url"] = f"https://{resource_prefix}syws.dev.azuresynapse.net"

    if not config["spark_config"]["azure_synapse"].get("workspace_dir"):
        config["spark_config"]["azure_synapse"]["workspace_dir"] =\
            f"abfss://{resource_prefix}fs@{resource_prefix}dls.dfs.core.windows.net/{project_name}"

    for k, v in DEFAULT_AZURE_SYNAPSE_SPARK_POOL_CONFIG.items():
        if not config["spark_config"]["azure_synapse"].get(k):
            config["spark_config"]["azure_synapse"][k] = v


def _set_databricks_config(
    config: Dict,
    project_name: str,
    cluster_id: str = None,
):
    """Set configs for Databricks spark cluster."""

    config["spark_config"]["databricks"] = config["spark_config"].get("databricks", {})

    if not config["spark_config"]["databricks"].get("work_dir"):
        config["spark_config"]["databricks"]["work_dir"] = f"dbfs:/{project_name}"

    if not config["spark_config"]["databricks"].get("config_template"):
        databricks_config = {
            "run_name": "FEATHR_FILL_IN",
            "libraries": [{"jar": "FEATHR_FILL_IN"}],
            "spark_jar_task": {
                "main_class_name": "FEATHR_FILL_IN",
                "parameters": ["FEATHR_FILL_IN"],
            },
        }
        if cluster_id is None:
            databricks_config["new_cluster"] = DEFAULT_DATABRICKS_CLUSTER_CONFIG
        else:
            databricks_config["existing_cluster_id"] = cluster_id

        config["spark_config"]["databricks"]["config_template"] = json.dumps(databricks_config)


def _config_kwargs_to_dict(**kwargs) -> Dict:
    """Parse config's keyword arguments to dictionary.
    e.g. `spark_config__spark_cluster="local"` will be parsed to `{"spark_config": {"spark_cluster": "local"}}`.
    """
    config = dict()

    for conf_key, conf_value in kwargs.items():
        if conf_value is None:
            continue

        conf = config
        keys = conf_key.split("__")
        for k in keys[:-1]:
            if k not in conf:
                conf[k] = dict()
            conf = conf[k]
        conf[keys[-1]] = conf_value

    return config


def _update_config(config: Dict, new_config: Dict):
    """Update config dictionary with the values in `new_config`."""
    for k, v in new_config.items():
        if k in config and isinstance(v, collections.abc.Mapping):
            _update_config(config[k], v)
        else:
            config[k] = v


def _verify_config(config: Dict):
    """Verify config."""
    if config["spark_config"]["spark_cluster"] == "azure_synapse":
        if not os.environ.get("ADLS_KEY"):
            raise ValueError("ADLS_KEY must be set in environment variables")
        elif (
            not os.environ.get("SPARK_CONFIG__AZURE_SYNAPSE__DEV_URL") and
            config["spark_config"]["azure_synapse"].get("dev_url") is None
        ):
            raise ValueError("Azure Synapse dev endpoint is not provided.")
        elif (
            not os.environ.get("SPARK_CONFIG__AZURE_SYNAPSE__POOL_NAME") and
            config["spark_config"]["azure_synapse"].get("pool_name") is None
        ):
            raise ValueError("Azure Synapse pool name is not provided.")

    elif config["spark_config"]["spark_cluster"] == "databricks":
        if not os.environ.get("DATABRICKS_WORKSPACE_TOKEN_VALUE"):
            raise ValueError("Databricks workspace token is not provided.")
        elif (
            not os.environ.get("SPARK_CONFIG__DATABRICKS__WORKSPACE_INSTANCE_URL") and
            config["spark_config"]["databricks"].get("workspace_instance_url") is None
        ):
            raise ValueError("Databricks workspace url is not provided.")


def _maybe_update_config_with_env_var(config: Dict, env_var_name: str):
    """Update config dictionary with the values in environment variables.
    e.g. `SPARK_CONFIG__SPARK_CLUSTER` will be parsed to `{"spark_config": {"spark_cluster": "local"}}`.
    """
    if not os.environ.get(env_var_name):
        return

    keys = env_var_name.lower().split("__")
    conf = config
    for k in keys[:-1]:
        if k not in conf:
            conf[k] = dict()
        conf = conf[k]

    conf[keys[-1]] = os.environ[env_var_name]



from feathr.definition.dtype import *
from feathr.registry.registry_utils import *
from feathr.utils._file_utils import write_to_file
from feathr.definition.anchor import FeatureAnchor
from feathr.constants import *
from feathr.definition.feature import Feature, FeatureType,FeatureBase
from feathr.definition.feature_derivations import DerivedFeature
from feathr.definition.repo_definitions import RepoDefinitions
from feathr.definition.source import HdfsSource, InputContext, JdbcSource, Source
from feathr.definition.transformation import (ExpressionTransformation, Transformation,
                                   WindowAggTransformation)
from feathr.definition.typed_key import TypedKey
from feathr.registry.feature_registry import FeathrRegistry
from feathr.definition.repo_definitions import RepoDefinitions
from pathlib import Path
from jinja2 import Template
import sys
from feathr.utils._file_utils import write_to_file
import importlib
import os

class FeathrConfigHelper(object):
    def __init__(self) -> None:
        pass
    def _get_py_files(self, path: Path) -> List[Path]:
        """Get all Python files under path recursively, excluding __init__.py"""
        py_files = []
        for item in path.glob('**/*.py'):
            if "__init__.py" != item.name:
                py_files.append(item)
        return py_files

    def _convert_to_module_path(self, path: Path, workspace_path: Path) -> str:
        """Convert a Python file path to its module path so that we can import it later"""
        prefix = os.path.commonprefix(
            [path.resolve(), workspace_path.resolve()])
        resolved_path = str(path.resolve())
        module_path = resolved_path[len(prefix): -len(".py")]
        # Convert features under nested folder to module name
        # e.g. /path/to/pyfile will become path.to.pyfile
        return (
            module_path
            .lstrip('/')
            .replace("/", ".")
        )

    def _extract_features_from_context(self, anchor_list, derived_feature_list, result_path: Path) -> RepoDefinitions:
        """Collect feature definitions from the context instead of python files"""
        definitions = RepoDefinitions(
            sources=set(),
            features=set(),
            transformations=set(),
            feature_anchors=set(),
            derived_features=set()
        )
        for derived_feature in derived_feature_list:
            if isinstance(derived_feature, DerivedFeature):
                definitions.derived_features.add(derived_feature)
                definitions.transformations.add(
                    vars(derived_feature)["transform"])
            else:
                raise RuntimeError(f"Please make sure you pass a list of `DerivedFeature` objects to the `derived_feature_list` argument. {str(type(derived_feature))} is detected.")

        for anchor in anchor_list:
            # obj is `FeatureAnchor`
            definitions.feature_anchors.add(anchor)
            # add the source section of this `FeatureAnchor` object
            definitions.sources.add(vars(anchor)['source'])
            for feature in vars(anchor)['features']:
                # get the transformation object from `Feature` or `DerivedFeature`
                if isinstance(feature, Feature):
                    # feature is of type `Feature`
                    definitions.features.add(feature)
                    definitions.transformations.add(vars(feature)["transform"])
                else:
                    
                    raise RuntimeError(f"Please make sure you pass a list of `Feature` objects. {str(type(feature))} is detected.")

        return definitions

    def _extract_features(self, workspace_path: Path) -> RepoDefinitions:
        """Collect feature definitions from the python file, convert them into feature config and save them locally"""
        os.chdir(workspace_path)
        # Add workspace path to system path so that we can load features defined in Python via import_module
        sys.path.append(str(workspace_path))
        definitions = RepoDefinitions(
            sources=set(),
            features=set(),
            transformations=set(),
            feature_anchors=set(),
            derived_features=set()
        )
        for py_file in self._get_py_files(workspace_path):
            module_path = self._convert_to_module_path(py_file, workspace_path)
            module = importlib.import_module(module_path)
            for attr_name in dir(module):
                obj = getattr(module, attr_name)
                if isinstance(obj, Source):
                    definitions.sources.add(obj)
                elif isinstance(obj, Feature):
                    definitions.features.add(obj)
                elif isinstance(obj, DerivedFeature):
                    definitions.derived_features.add(obj)
                elif isinstance(obj, FeatureAnchor):
                    definitions.feature_anchors.add(obj)
                elif isinstance(obj, Transformation):
                    definitions.transformations.add(obj)
        return definitions

    def save_to_feature_config(self, workspace_path: Path, config_save_dir: Path):
        """Save feature definition within the workspace into HOCON feature config files"""
        repo_definitions = self._extract_features(workspace_path)
        self._save_request_feature_config(repo_definitions, config_save_dir)
        self._save_anchored_feature_config(repo_definitions, config_save_dir)
        self._save_derived_feature_config(repo_definitions, config_save_dir)

    def save_to_feature_config_from_context(self, anchor_list, derived_feature_list, local_workspace_dir: Path):
        """Save feature definition within the workspace into HOCON feature config files from current context, rather than reading from python files"""
        repo_definitions = self._extract_features_from_context(
            anchor_list, derived_feature_list, local_workspace_dir)
        self._save_request_feature_config(repo_definitions, local_workspace_dir)
        self._save_anchored_feature_config(repo_definitions, local_workspace_dir)
        self._save_derived_feature_config(repo_definitions, local_workspace_dir)

    def _save_request_feature_config(self, repo_definitions: RepoDefinitions, local_workspace_dir="./"):
        config_file_name = "feature_conf/auto_generated_request_features.conf"
        tm = Template(
            """
// THIS FILE IS AUTO GENERATED. PLEASE DO NOT EDIT.
anchors: {
    {% for anchor in feature_anchors %}
        {% if anchor.source.name == "PASSTHROUGH" %}
            {{anchor.to_feature_config()}}
        {% endif %}
    {% endfor %}
}
"""
        )

        request_feature_configs = tm.render(
            feature_anchors=repo_definitions.feature_anchors)
        config_file_path = os.path.join(local_workspace_dir, config_file_name)
        write_to_file(content=request_feature_configs,
                      full_file_name=config_file_path)

    @classmethod
    def _save_anchored_feature_config(self, repo_definitions: RepoDefinitions, local_workspace_dir="./"):
        config_file_name = "feature_conf/auto_generated_anchored_features.conf"
        tm = Template(
            """
// THIS FILE IS AUTO GENERATED. PLEASE DO NOT EDIT.
anchors: {
    {% for anchor in feature_anchors %}
        {% if not anchor.source.name == "PASSTHROUGH" %}
            {{anchor.to_feature_config()}}
        {% endif %}
    {% endfor %}
}

sources: {
    {% for source in sources%}
        {% if not source.name == "PASSTHROUGH" %}
            {{source.to_feature_config()}}
        {% endif %}
    {% endfor %}
}
"""
        )
        anchored_feature_configs = tm.render(feature_anchors=repo_definitions.feature_anchors,
                                             sources=repo_definitions.sources)
        config_file_path = os.path.join(local_workspace_dir, config_file_name)
        write_to_file(content=anchored_feature_configs,
                      full_file_name=config_file_path)

    @classmethod
    def _save_derived_feature_config(self, repo_definitions: RepoDefinitions, local_workspace_dir="./"):
        config_file_name = "feature_conf/auto_generated_derived_features.conf"
        tm = Template(
            """
anchors: {}
derivations: {
    {% for derived_feature in derived_features %}
        {{derived_feature.to_feature_config()}}
    {% endfor %}
}
"""
        )
        derived_feature_configs = tm.render(
            derived_features=repo_definitions.derived_features)
        config_file_path = os.path.join(local_workspace_dir, config_file_name)
        write_to_file(content=derived_feature_configs,
                      full_file_name=config_file_path)
                      



from copy import deepcopy
import json
import os
import pathlib
import re
import time
import urllib.request
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple
from urllib.parse import urlparse
from os.path import basename
from enum import Enum
import tempfile
from azure.identity import (ChainedTokenCredential, DefaultAzureCredential,
                            DeviceCodeCredential, EnvironmentCredential,
                            ManagedIdentityCredential)
from azure.storage.filedatalake import DataLakeServiceClient, DataLakeDirectoryClient
from azure.synapse.spark import SparkClient
from azure.synapse.spark.models import SparkBatchJobOptions
from loguru import logger
from requests import request
from tqdm import tqdm

from feathr.spark_provider._abc import SparkJobLauncher
from feathr.constants import *
from feathr.version import get_maven_artifact_fullname

class LivyStates(Enum):
    """ Adapt LivyStates over to relax the dependency for azure-synapse-spark pacakge.
    Definition is here:
    https://github.com/Azure/azure-sdk-for-python/blob/main/sdk/synapse/azure-synapse-spark/azure/synapse/spark/models/_spark_client_enums.py#L38
    """

    NOT_STARTED = "not_started"
    STARTING = "starting"
    IDLE = "idle"
    BUSY = "busy"
    SHUTTING_DOWN = "shutting_down"
    ERROR = "error"
    DEAD = "dead"
    KILLED = "killed"
    SUCCESS = "success"
    RUNNING = "running"
    RECOVERING = "recovering"


class _FeathrSynapseJobLauncher(SparkJobLauncher):
    """
    Submits spark jobs to a Synapse spark cluster.
    """

    def __init__(self, synapse_dev_url: str, pool_name: str, datalake_dir: str, executor_size: str, executors: int, credential=None):
        # use DeviceCodeCredential if EnvironmentCredential is not available
        self.credential = credential
        # use the same credential for authentication to avoid further login.
        self._api = _SynapseJobRunner(
            synapse_dev_url, pool_name, executor_size=executor_size, executors=executors, credential=self.credential)
        self._datalake = _DataLakeFiler(
            datalake_dir, credential=self.credential)
        # Save Synapse parameters to retrieve driver log
        self._synapse_dev_url = synapse_dev_url
        self._pool_name = pool_name

    def upload_or_get_cloud_path(self, local_path_or_cloud_src_path: str, tar_dir_path: Optional[str] = None):
        """
        Supports transferring file from an http path to cloud working storage, or upload directly from a local storage,
        or copying files from a source datalake directory to a target datalake directory
        """
        if local_path_or_cloud_src_path.startswith('abfs') or local_path_or_cloud_src_path.startswith('wasb'):
            if tar_dir_path is None or not (tar_dir_path.startswith('abfs') or tar_dir_path.startswith('wasb')):
                raise RuntimeError(
                f"Failed to copy files from dbfs directory: {local_path_or_cloud_src_path}. {tar_dir_path} is not a valid target directory path"
            )
            [_, source_exist] = self._datalake._dir_exists(local_path_or_cloud_src_path)
            if not source_exist:
                raise RuntimeError(f"Source folder:{local_path_or_cloud_src_path} doesn't exist. Please make sure it's a valid path")
            [dir_client, target_exist] = self._datalake._dir_exists(tar_dir_path)
            if target_exist:
                logger.warning('Target cloud directory {} already exists. Please use another one.', tar_dir_path)
                return tar_dir_path
            dir_client.create_directory()
            tem_dir_obj = tempfile.TemporaryDirectory()
            self._datalake.download_file(local_path_or_cloud_src_path, tem_dir_obj.name)
            self._datalake.upload_file_to_workdir(tem_dir_obj.name, tar_dir_path, dir_client)
            logger.info('{} is uploaded to location: {}',
                    local_path_or_cloud_src_path, tar_dir_path)
            return tar_dir_path
                            
        logger.info('Uploading {} to cloud..', local_path_or_cloud_src_path)
        res_path = self._datalake.upload_file_to_workdir(
            local_path_or_cloud_src_path)

        logger.info('{} is uploaded to location: {}',
                    local_path_or_cloud_src_path, res_path)
        return res_path

    def download_result(self, result_path: str, local_folder: str, is_file_path: bool = False):
        """
        Supports downloading files from the result folder
        """
        if is_file_path:
            paths = result_path.rsplit('/',1)
            if len(paths) != 2:
                raise RuntimeError(f"Invalid single file path: {result_path}")
            return self._datalake.download_file(paths[0]+'/', local_folder, paths[1])
        return self._datalake.download_file(result_path, local_folder, None)
    
    
    def cloud_dir_exists(self, dir_path: str) -> bool:
        """
        Checks if a directory already exists in the datalake
        """
        
        [_, exists] = self._datalake._dir_exists(dir_path)
        return exists

    def submit_feathr_job(self, job_name: str, main_jar_path: str = None,  main_class_name: str = None, arguments: List[str] = None,
                          python_files: List[str]= None, reference_files_path: List[str] = None, job_tags: Dict[str, str] = None,
                          configuration: Dict[str, str] = {}, properties: Dict[str, str] = {}):
        """
        Submits the feathr job
        Refer to the Apache Livy doc for more details on the meaning of the parameters:
        https://livy.apache.org/docs/latest/rest-api.html

        reference files: put everything there and the function will automatically categorize them based on the
        extension name to either the "files" argument in the Livy API, or the "jars" argument in the Livy API. The
        path can be local path and this function will automatically upload the function to the corresponding azure
        storage

        Also, note that the Spark application will automatically run on YARN cluster mode. You cannot change it if
        you are running with Azure Synapse.

        Args:
            job_name (str): name of the job
            main_jar_path (str): main file paths, usually your main jar file
            main_class_name (str): name of your main class
            arguments (str): all the arguments you want to pass into the spark job
            job_tags (str): tags of the job, for example you might want to put your user ID, or a tag with a certain information
            configuration (Dict[str, str]): Additional configs for the spark job
            properties (Dict[str, str]): Additional System Properties for the spark job
        """

        if properties:
            arguments.append("--system-properties=%s" % json.dumps(properties))

        if configuration:
            cfg = configuration.copy()  # We don't want to mess up input parameters
        else:
            cfg = {}
        if not main_jar_path:
            # We don't have the main jar, use Maven
            # Add Maven dependency to the job configuration
            logger.info(f"Main JAR file is not set, using default package '{get_maven_artifact_fullname()}' from Maven")
            if "spark.jars.packages" in cfg:
                cfg["spark.jars.packages"] = ",".join(
                    [cfg["spark.jars.packages"], get_maven_artifact_fullname()])
            else:
                cfg["spark.jars.packages"] = get_maven_artifact_fullname()

            if not python_files:
                # This is a JAR job
                # Azure Synapse/Livy doesn't allow JAR job starts from Maven directly, we must have a jar file uploaded.
                # so we have to use a dummy jar as the main file.
                # Use the no-op jar as the main file
                # This is a dummy jar which contains only one `org.example.Noop` class with one empty `main` function which does nothing
                current_dir = pathlib.Path(__file__).parent.resolve()
                main_jar_path = os.path.join(current_dir, "noop-1.0.jar")
            else:
                # This is a PySpark job, no more things to do
                pass
        main_jar_cloud_path = None
        if main_jar_path:
            # Now we have a main jar, either feathr or noop
            if main_jar_path.startswith('abfs'):
                main_jar_cloud_path = main_jar_path
                logger.info(
                    'Cloud path {} is used for running the job: {}', main_jar_path, job_name)
            else:
                logger.info('Uploading jar from {} to cloud for running job: {}',
                            main_jar_path, job_name)
                main_jar_cloud_path = self._datalake.upload_file_to_workdir(main_jar_path)
                logger.info('{} is uploaded to {} for running job: {}',
                            main_jar_path, main_jar_cloud_path, job_name)
        else:
            # We don't have the main Jar, and this is a PySpark job so we don't use `noop.jar` either
            # Keep `main_jar_cloud_path` as `None` as we already added maven package into cfg
            pass

        reference_file_paths = []
        for file_path in reference_files_path:
            reference_file_paths.append(
                self._datalake.upload_file_to_workdir(file_path))

        self.current_job_info = self._api.create_spark_batch_job(job_name=job_name,
                                                                 main_file=main_jar_cloud_path,
                                                                 class_name=main_class_name,
                                                                 python_files=python_files,
                                                                 arguments=arguments,
                                                                 reference_files=reference_files_path,
                                                                 tags=job_tags,
                                                                 configuration=cfg)
        logger.info('See submitted job here: https://web.azuresynapse.net/en-us/monitoring/sparkapplication')
        return self.current_job_info

    def wait_for_completion(self, timeout_seconds: Optional[float]) -> bool:
        """
        Returns true if the job completed successfully
        """          
        start_time = time.time()
        while (timeout_seconds is None) or (time.time() - start_time < timeout_seconds):
            status = self.get_status()
            logger.info('Current Spark job status: {}', status)
            if status in {LivyStates.SUCCESS.value}:
                return True
            elif status in {LivyStates.ERROR.value, LivyStates.DEAD.value, LivyStates.KILLED.value}:
                logger.error("Feathr job has failed.")
                error_msg = self._api.get_driver_log(self.current_job_info.id).decode('utf-8')
                logger.error(error_msg)
                logger.error("The size of the whole error log is: {}. The logs might be truncated in some cases (such as in Visual Studio Code) so only the top a few lines of the error message is displayed. If you cannot see the whole log, you may want to extend the setting for output size limit.", len(error_msg))
                return False
            else:
                time.sleep(30)
        else:
            raise TimeoutError('Timeout waiting for job to complete')

    def get_status(self) -> str:
        """Get current job status

        Returns:
            str: Status of the current job
        """
        job = self._api.get_spark_batch_job(self.current_job_info.id)
        assert job is not None
        return job.state

    def get_job_result_uri(self) -> str:
        """Get job output uri

        Returns:
            str: `output_path` field in the job tags
        """
        tags = self._api.get_spark_batch_job(self.current_job_info.id).tags
        # in case users call this API even when there's no tags available
        return None if tags is None else tags[OUTPUT_PATH_TAG]

    def get_job_tags(self) -> Dict[str, str]:
        """Get job tags

        Returns:
            Dict[str, str]: a dict of job tags
        """
        return self._api.get_spark_batch_job(self.current_job_info.id).tags

class _SynapseJobRunner(object):
    """
    Class to interact with Synapse Spark cluster
    """
    def __init__(self, synapse_dev_url, spark_pool_name, credential=None, executor_size='Small', executors=2):
        self._synapse_dev_url = synapse_dev_url
        self._spark_pool_name = spark_pool_name
        if credential is None:
            logger.warning('No valid Azure credential detected. Using DefaultAzureCredential')
            credential = DefaultAzureCredential()
        self._credential = credential

        self.client = SparkClient(
            credential=credential,
            endpoint=synapse_dev_url,
            spark_pool_name=spark_pool_name
        )

        self._executor_size = executor_size
        self._executors = executors
        self.EXECUTOR_SIZE = {'Small': {'Cores': 4, 'Memory': '28g'}, 'Medium': {'Cores': 8, 'Memory': '56g'},
                              'Large': {'Cores': 16, 'Memory': '112g'}}

    def _categorized_files(self, reference_files: List[str]):
        """categorize files to make sure they are in the ready to submissio format

        Args:
            reference_files (List[str]): a list of reference files, can be either jars, pyfiles, etc.

        Returns:
            a tuple with all the files and jars clasified into list
        """
        if reference_files == None:
            return None, None

        files = []
        jars = []
        for file in reference_files:
            file = file.strip()
            if file.endswith('.jar'):
                jars.append(file)
            else:
                files.append(file)
        return files, jars

    def get_spark_batch_job(self, job_id:int):
        """
        Get the job object by searching a certain ID
        """

        return self.client.spark_batch.get_spark_batch_job(job_id, detailed=True)

    def get_spark_batch_jobs(self):
        """
        Get all the jobs in a certain Spark pool
        """

        return self.client.spark_batch.get_spark_batch_jobs(detailed=True)

    def cancel_spark_batch_job(self, job_id:int):
        """
        Cancel a job by searching a certain ID
        """

        return self.client.spark_batch.cancel_spark_batch_job(job_id)

    def create_spark_batch_job(self, job_name, main_file, class_name=None,
                               arguments=None, python_files=None, reference_files=None, archives=None, configuration=None, tags=None):
        """
        Submit a spark job to a certain cluster
        """

        files, jars = self._categorized_files(reference_files)
        driver_cores = self.EXECUTOR_SIZE[self._executor_size]['Cores']
        driver_memory = self.EXECUTOR_SIZE[self._executor_size]['Memory']
        executor_cores = self.EXECUTOR_SIZE[self._executor_size]['Cores']
        executor_memory = self.EXECUTOR_SIZE[self._executor_size]['Memory']

        # If we have a main jar, it needs to be added as dependencies for pyspark job
        # Otherwise it's a PySpark job with Feathr JAR from Maven
        if main_file:
            jars = jars + [main_file]
        elif not python_files:
            # These 2 parameters should not be empty at the same time
            raise ValueError("Main JAR is not set for the Spark job")

        # If file=main_file, then it's using only Scala Spark
        # If file=python_files[0], then it's using Pyspark
        spark_execution_file = python_files[0] if python_files else main_file

        spark_batch_job_options = SparkBatchJobOptions(
            tags=tags,
            name=job_name,
            file=spark_execution_file,
            class_name=class_name,
            python_files=python_files[1:],
            arguments=arguments,
            jars=jars,
            files=files,
            archives=archives,
            configuration=configuration,
            driver_memory=driver_memory,
            driver_cores=driver_cores,
            executor_memory=executor_memory,
            executor_cores=executor_cores,
            executor_count=self._executors)

        return self.client.spark_batch.create_spark_batch_job(spark_batch_job_options, detailed=True)

    def get_driver_log(self, job_id) -> str:
        # @see: https://docs.microsoft.com/en-us/azure/synapse-analytics/spark/connect-monitor-azure-synapse-spark-application-level-metrics
        app_id = self.get_spark_batch_job(job_id).app_id
        url = "%s/sparkhistory/api/v1/sparkpools/%s/livyid/%s/applications/%s/driverlog/stderr/?isDownload=true" % (self._synapse_dev_url, self._spark_pool_name, job_id, app_id)
        token = self._credential.get_token("https://dev.azuresynapse.net/.default").token
        req = urllib.request.Request(url=url, headers={"authorization": "Bearer %s" % token})
        resp = urllib.request.urlopen(req)
        return resp.read()


class _DataLakeFiler(object):
    """
    Class to interact with Azure Data Lake Storage.
    """
    def __init__(self, datalake_dir, credential=None):
        # A datalake path would be something like this:
        # "abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/frame_getting_started" after this
        # split, datalake_path_split should give out something like this; ['abfss:', '', 'feathrazuretest3fs',
        # 'feathrazuretest3storage.dfs.core.windows.net', 'frame_getting_started'] datalake_path_split[0] should be
        # the protocal, datalake_path_split[1] is empty, datalake_path_split[2] is the file system used (aka
        # container name), datalake_path_split[3] should be the full name of this target path,
        # and datalake_path_split[3:] would be all the directory in this particular container split datalake names by
        # "/" or "@"
        datalake_path_split = list(filter(None, re.split('/|@', datalake_dir)))
        assert len(datalake_path_split) >= 3

        if credential is None:
            raise RuntimeError("Invalid credential provided.")

        account_url = "https://" + datalake_path_split[2]

        self.file_system_client = DataLakeServiceClient(
            credential=credential,
            account_url=account_url
        ).get_file_system_client(datalake_path_split[1])

        if len(datalake_path_split) > 3:
            # directory exists in datalake path
            self.dir_client = self.file_system_client.get_directory_client(
                '/'.join(datalake_path_split[3:]))
            self.dir_client.create_directory()
        else:
            # otherwise use root folder instead
            self.dir_client = self.file_system_client.get_directory_client('/')

        self.datalake_dir = datalake_dir + \
            '/' if datalake_dir[-1] != '/' else datalake_dir

    def upload_file_to_workdir(self, src_file_path: str, tar_dir_path: Optional[str] = "", tar_dir_client: Optional[DataLakeDirectoryClient] = None) -> str:
        """
        Handles file upload to the corresponding datalake storage. If a path starts with "wasb" or "abfs",
        it will skip uploading and return the original path; otherwise it will upload the source file to the working
        dir
        """

        src_parse_result = urlparse(src_file_path)
        if src_parse_result.scheme.startswith('http'):
            file_name = basename(src_file_path)
            file_client = self.dir_client.create_file(file_name)
            # returned paths for the uploaded file
            returned_path = self.datalake_dir + file_name
            with urllib.request.urlopen(src_file_path) as f:
                data = f.read()
                file_client.upload_data(data, overwrite=True)
                logger.info("{} is downloaded and then uploaded to location: {}", src_file_path, returned_path)
        elif src_parse_result.scheme.startswith('abfs') or src_parse_result.scheme.startswith('wasb'):
            # passed a cloud path
            logger.info("Skip uploading file {} as it's already in the cloud", src_file_path)
            returned_path = src_file_path
        else:
            # else it should be a local file path or dir
            if os.path.isdir(src_file_path):
                logger.info("Uploading folder {}", src_file_path)
                dest_paths = []
                if tar_dir_client is not None:
                    # Only supports uploading local files/dir to datalake dir for now
                     for item in Path(src_file_path).iterdir():
                        returned_path = self.upload_file(item.resolve(), tar_dir_path, tar_dir_client)
                        dest_paths.extend([returned_path]) 
                else:
                    for item in Path(src_file_path).glob('**/*.conf'):
                        returned_path = self.upload_file(item.resolve())
                        dest_paths.extend([returned_path])
                returned_path = ','.join(dest_paths)
            else:
                returned_path = self.upload_file(src_file_path)
        return returned_path

    def upload_file(self, src_file_path, tar_dir_path: Optional[str]="", tar_dir_client: Optional[DataLakeDirectoryClient] = None)-> str:
        file_name = basename(src_file_path)
        logger.info("Uploading file {}", file_name)
        # TODO: add handling for only tar_dir_client or tar_dir_path is provided
        file_client = self.dir_client.create_file(file_name) if tar_dir_client is None else tar_dir_client.create_file(file_name)
        returned_path = self.datalake_dir + file_name if tar_dir_path == "" else tar_dir_path + file_name
        with open(src_file_path, 'rb') as f:
            data = f.read()
            file_client.upload_data(data, overwrite=True)
        logger.info("{} is uploaded to location: {}", src_file_path, returned_path)
        return returned_path
        

    def download_file(self, target_adls_directory: str, local_dir_cache: str, file_name: str = None):
        """
        Download file to a local cache. Supporting download a folder and the content in its subfolder.
        Note that the code will just download the content in the root folder, and the folder in the next level (rather than recursively for all layers of folders)

        Args:
            target_adls_directory (str): target ADLS directory
            local_dir_cache (str): local cache to store local results
            file_name (str): only download the file with name 'file_name' under the target directory if it's provided (default as None)
        """
        logger.info('Beginning reading of results from {}',
                    target_adls_directory)
        parse_result = urlparse(target_adls_directory)
        if parse_result.path == '':
            parse_result.path = '/'
        directory_client = self.file_system_client.get_directory_client(
            parse_result.path)
        
        if file_name is not None:
            local_paths = [os.path.join(local_dir_cache, file_name)]
            self._download_file_list(local_paths, [file_name], directory_client)
            logger.info('Finish downloading file {} from {} to {}.',
                    file_name, target_adls_directory, local_dir_cache)
            return
        
        # returns the paths to all the files in the target director in ADLS
        # get all the paths that are not under a directory
        result_paths = [basename(file_path.name) for file_path in self.file_system_client.get_paths(
            path=parse_result.path, recursive=False) if not file_path.is_directory]

        # get all the paths that are directories and download them
        result_folders = [file_path.name for file_path in self.file_system_client.get_paths(
            path=parse_result.path) if file_path.is_directory]

        # list all the files under the certain folder, and download them preserving the hierarchy
        for folder in result_folders:
            folder_name = basename(folder)
            file_in_folder = [os.path.join(folder_name, basename(file_path.name)) for file_path in self.file_system_client.get_paths(
                path=folder, recursive=False) if not file_path.is_directory]
            local_paths = [os.path.join(local_dir_cache, file_name)
                       for file_name in file_in_folder]
            self._download_file_list(local_paths, file_in_folder, directory_client)

        # download files that are in the result folder
        local_paths = [os.path.join(local_dir_cache, file_name)
                       for file_name in result_paths]
        self._download_file_list(local_paths, result_paths, directory_client)

        logger.info('Finish downloading files from {} to {}.',
                    target_adls_directory, local_dir_cache)
        
    def _download_file_list(self, local_paths: List[str], result_paths, directory_client):
        '''
        Download filelist to local
        '''
        for idx, file_to_write in enumerate(tqdm(result_paths,desc="Downloading result files: ")):
            try:
                os.makedirs(os.path.dirname(local_paths[idx]), exist_ok=True)
                local_file = open(local_paths[idx], 'wb')
                file_client = directory_client.get_file_client(file_to_write)
                download = file_client.download_file()
                downloaded_bytes = download.readall()
                local_file.write(downloaded_bytes)
                local_file.close()               
            except Exception as e:
                logger.error(e)       
                
    def _dir_exists(self, dir_path:str):
        '''
        Check if a directory in datalake already exists. Will also return the directory client
        '''
        datalake_path_split = list(filter(None, re.split('/|@', dir_path)))
        if len(datalake_path_split) <= 3:
            raise RuntimeError("Invalid directory path for datalake: {dir_path}")
        dir_client = self.file_system_client.get_directory_client(
                '/'.join(datalake_path_split[3:]))
        return [dir_client, dir_client.exists()]
        


import os
from datetime import datetime
from pathlib import Path

import pytest

from feathr import BOOLEAN, FLOAT, INT32
from feathr import FeathrClient
from feathr import FeatureQuery
from feathr import JdbcSource
from feathr import ObservationSettings
from feathr import TypedKey
from feathr import ValueType, Feature, FeatureAnchor, DerivedFeature, WindowAggTransformation, INPUT_CONTEXT
from feathr.utils.job_utils import get_result_df
from test_fixture import get_online_test_table_name
from test_utils.constants import Constants

def basic_test_setup(config_path: str):
    """
    Basically this is same as the one in `text_fixture.py` with the same name.
    The difference is the `batch_source` is configured to read from JDBC instead of HDFS
    """

    client = FeathrClient(config_path=config_path)

    # Using database under @windoze account, so this e2e test still doesn't work in CI
    batch_source = JdbcSource(name="nycTaxiBatchJdbcSource",
                              url="jdbc:sqlserver://feathrtestsql4.database.windows.net:1433;database=testsql;encrypt=true;trustServerCertificate=false;hostNameInCertificate=*.database.windows.net;loginTimeout=30;",
                              dbtable="green_tripdata_2020_04",
                              auth="USERPASS",
                              event_timestamp_column="lpep_dropoff_datetime",
                              timestamp_format="yyyy-MM-dd HH:mm:ss")


    f_trip_distance = Feature(name="f_trip_distance",
                              feature_type=FLOAT, transform="trip_distance")
    f_trip_time_duration = Feature(name="f_trip_time_duration",
                                   feature_type=INT32,
                                   transform="(to_unix_timestamp(lpep_dropoff_datetime) - to_unix_timestamp(lpep_pickup_datetime))/60")

    features = [
        f_trip_distance,
        f_trip_time_duration,
        Feature(name="f_is_long_trip_distance",
                feature_type=BOOLEAN,
                transform="cast_float(trip_distance)>30"),
        Feature(name="f_day_of_week",
                feature_type=INT32,
                transform="dayofweek(lpep_dropoff_datetime)"),
    ]


    request_anchor = FeatureAnchor(name="request_features",
                                   source=INPUT_CONTEXT,
                                   features=features)

    f_trip_time_distance = DerivedFeature(name="f_trip_time_distance",
                                          feature_type=FLOAT,
                                          input_features=[
                                              f_trip_distance, f_trip_time_duration],
                                          transform="f_trip_distance * f_trip_time_duration")

    f_trip_time_rounded = DerivedFeature(name="f_trip_time_rounded",
                                         feature_type=INT32,
                                         input_features=[f_trip_time_duration],
                                         transform="f_trip_time_duration % 10")

    location_id = TypedKey(key_column="DOLocationID",
                           key_column_type=ValueType.INT32,
                           description="location id in NYC",
                           full_name="nyc_taxi.location_id")
    
    # This feature is read from Jdbc data source
    agg_features = [Feature(name="f_location_avg_fare",
                            key=location_id,
                            feature_type=FLOAT,
                            transform=WindowAggTransformation(agg_expr="cast_float(fare_amount)",
                                                              agg_func="AVG",
                                                              window="90d")),
                    ]

    agg_anchor = FeatureAnchor(name="aggregationFeatures",
                               source=batch_source,
                               features=agg_features)

    client.build_features(anchor_list=[agg_anchor, request_anchor], derived_feature_list=[
        f_trip_time_distance, f_trip_time_rounded])

    return client

@pytest.mark.skip(reason="Requires database with test data imported, which doesn't exist in the current CI env")
def test_feathr_get_offline_features():
    """
    Test FeathrClient() get_offline_features works with Jdbc data source.
    Currently it doesn't work in CI env hence being marked as `skip`,
    because it requires a database with `green_tripdata_2020-04.csv` imported into a table named `green_tripdata_2020_04`.
    To run this test, you'll also need following environment variables set correctly:
    - nycTaxiBatchJdbcSource_USER: The user name to login to database server
    - nycTaxiBatchJdbcSource_PASSWORD: The password to login to database server
    These 2 variables will be passed to the Spark job in `--system-properties` parameter so Spark can access the database
    """

    test_workspace_dir = Path(
        __file__).parent.resolve() / "test_user_workspace"
    # os.chdir(test_workspace_dir)

    client = basic_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))

    location_id = TypedKey(key_column="DOLocationID",
                            key_column_type=ValueType.INT32,
                            description="location id in NYC",
                            full_name="nyc_taxi.location_id")

    feature_query = FeatureQuery(
        feature_list=["f_location_avg_fare"], key=location_id)
    settings = ObservationSettings(
        observation_path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04.csv",
        event_timestamp_column="lpep_dropoff_datetime",
        timestamp_format="yyyy-MM-dd HH:mm:ss")

    now = datetime.now()
    output_path = ''.join(['dbfs:/feathrazure_cijob','_', str(now.minute), '_', str(now.second), ".avro"])
    
    client.get_offline_features(observation_settings=settings,
                                feature_query=feature_query,
                                output_path=output_path)

    # assuming the job can successfully run; otherwise it will throw exception
    client.wait_job_to_finish(timeout_sec=Constants.SPARK_JOB_TIMEOUT_SECONDS)

    # download result and just assert the returned result is not empty
    res_df = get_result_df(client)
    assert res_df.shape[0] > 0

    
if __name__ == "__main__":
    test_feathr_get_offline_features()


from azure.keyvault.secrets import SecretClient
from azure.identity import DefaultAzureCredential
from loguru import logger
from azure.core.exceptions import ResourceNotFoundError

class AzureKeyVaultClient:
    def __init__(self, akv_name: str):
        self.akv_name = akv_name
        self.secret_client = None

    def get_feathr_akv_secret(self, secret_name: str):
        """Get Feathr Secrets from Azure Key Vault. Note that this function will replace '_' in `secret_name` with '-' since Azure Key Vault doesn't support it

        Returns:
            _type_: _description_
        """
        if self.secret_client is None:
            self.secret_client = SecretClient(
                vault_url = f"https://{self.akv_name}.vault.azure.net",
                credential=DefaultAzureCredential()
            )
        try:
            # replace '_' with '-' since Azure Key Vault doesn't support it
            variable_replaced = secret_name.replace('_','-') #.upper()
            logger.info('Fetching the secret {} from Key Vault {}.', variable_replaced, self.akv_name)
            secret = self.secret_client.get_secret(variable_replaced)
            logger.info('Secret {} fetched from Key Vault {}.', variable_replaced, self.akv_name)
            return secret.value
        except ResourceNotFoundError as e:
            logger.error(f"Secret {secret_name} cannot be found in Key Vault {self.akv_name}.")
            raise


from pathlib import Path
from tempfile import TemporaryDirectory
from threading import local
from urllib.parse import urlparse

import pandas as pd
from pyspark.sql import DataFrame, SparkSession

from feathr.datasets import NYC_TAXI_SMALL_URL
from feathr.datasets.utils import maybe_download
from feathr.utils.platform import is_databricks


def get_pandas_df(
    local_cache_path: str = None,
) -> pd.DataFrame:
    """Get NYC taxi fare prediction data samples as a pandas DataFrame.

    Refs:
        https://www1.nyc.gov/site/tlc/about/tlc-trip-record-data.page

    Args:
        local_cache_path (optional): Local cache file path to download the data set.
            If local_cache_path is a directory, the source file name will be added.

    Returns:
        pandas DataFrame
    """
    # if local_cache_path params is not provided then create a temporary folder
    if local_cache_path is None:
        local_cache_path = TemporaryDirectory().name

    # If local_cache_path is a directory, add the source file name.
    src_filepath = Path(urlparse(NYC_TAXI_SMALL_URL).path)
    dst_path = Path(local_cache_path)
    if dst_path.suffix != src_filepath.suffix:
        local_cache_path = str(dst_path.joinpath(src_filepath.name))

    maybe_download(src_url=NYC_TAXI_SMALL_URL, dst_filepath=local_cache_path)

    pdf = pd.read_csv(local_cache_path)

    return pdf


def get_spark_df(
    spark: SparkSession,
    local_cache_path: str,
) -> DataFrame:
    """Get NYC taxi fare prediction data samples as a spark DataFrame.

    Refs:
        https://www1.nyc.gov/site/tlc/about/tlc-trip-record-data.page

    Args:
        spark: Spark session.
        local_cache_path: Local cache file path to download the data set.
            If local_cache_path is a directory, the source file name will be added.

    Returns:
        Spark DataFrame
    """
    # In spark, local_cache_path should be a persist directory or file path
    if local_cache_path is None:
        raise ValueError("In spark, `local_cache_path` should be a persist directory or file path.")

    # If local_cache_path is a directory, add the source file name.
    src_filepath = Path(urlparse(NYC_TAXI_SMALL_URL).path)
    dst_path = Path(local_cache_path)
    if dst_path.suffix != src_filepath.suffix:
        local_cache_path = str(dst_path.joinpath(src_filepath.name))

    # Databricks uses "dbfs:/" prefix for spark paths and "/dbfs/" prefix for python paths
    if is_databricks():
        # If local_cache_path is a python path, convert it to a spark path.
        if local_cache_path.startswith("/dbfs/"):
            local_cache_path = local_cache_path.replace("/dbfs", "dbfs:", 1)
        # If local_cache_path is not a spark path, add the spark path prefix.
        elif not local_cache_path.startswith("dbfs:"):
            local_cache_path = f"dbfs:/{local_cache_path.lstrip('/')}"

        python_local_cache_path = local_cache_path.replace("dbfs:", "/dbfs", 1)
    # TODO add "if is_synapse()"
    else:
        python_local_cache_path = local_cache_path

    maybe_download(src_url=NYC_TAXI_SMALL_URL, dst_filepath=python_local_cache_path)

    df = spark.read.option("header", True).csv(local_cache_path)

    return df



"""Dataset utilities
"""
import logging
import math
from pathlib import Path
import requests
from urllib.parse import urlparse

from tqdm import tqdm


log = logging.getLogger(__name__)


def maybe_download(src_url: str, dst_filepath: str, expected_bytes=None) -> bool:
    """Check if file exists. If not, download and return True. Else, return False.

    Refs:
        https://github.com/microsoft/recommenders/blob/main/recommenders/datasets/download_utils.py

    Args:
        src_url: Source file URL.
        dst_filepath: Destination file path.
        expected_bytes (optional): Expected bytes of the file to verify.

    Returns:
        bool: Whether the file was downloaded or not
    """
    dst_filepath = Path(dst_filepath)

    if dst_filepath.is_file():
        log.info(f"File {str(dst_filepath)} already exists")
        return False

    # Check dir if exists. If not, create one
    dst_filepath.parent.mkdir(parents=True, exist_ok=True)

    response = requests.get(src_url, stream=True)
    if response.status_code == 200:
        log.info(f"Downloading {src_url}")
        total_size = int(response.headers.get("content-length", 0))
        block_size = 1024
        num_iterables = math.ceil(total_size / block_size)
        with open(str(dst_filepath.resolve()), "wb") as file:
            for data in tqdm(
                response.iter_content(block_size),
                total=num_iterables,
                unit="KB",
                unit_scale=True,
            ):
                file.write(data)

        # Verify the file size
        if expected_bytes is not None and expected_bytes != dst_filepath.stat().st_size:
            # Delete the file since the size is not the same as the expected one.
            dst_filepath.unlink()
            raise IOError(f"Failed to verify {str(dst_filepath)}. Maybe interrupted while downloading?")
        else:
            return True

    else:
        response.raise_for_status()
        # If not HTTPError yet still cannot download
        raise Exception(f"Problem downloading {src_url}")



import base64
import copy
import json
import logging
import os
import tempfile
from typing import Any, Dict, List, Tuple, Union, Set

from azure.identity import DefaultAzureCredential
from jinja2 import Template
from loguru import logger
from pyhocon import ConfigFactory
import redis

from feathr.constants import *
from feathr.definition._materialization_utils import _to_materialization_config
from feathr.definition.anchor import FeatureAnchor
from feathr.definition.config_helper import FeathrConfigHelper
from feathr.definition.feature import FeatureBase
from feathr.definition.feature_derivations import DerivedFeature
from feathr.definition.materialization_settings import MaterializationSettings
from feathr.definition.monitoring_settings import MonitoringSettings
from feathr.definition.query_feature_list import FeatureQuery
from feathr.definition.settings import ObservationSettings, ConflictsAutoCorrection
from feathr.definition.sink import HdfsSink, Sink
from feathr.definition.source import InputContext
from feathr.definition.transformation import WindowAggTransformation
from feathr.definition.typed_key import TypedKey
from feathr.protobuf.featureValue_pb2 import FeatureValue
from feathr.registry._feathr_registry_client import _FeatureRegistry, derived_feature_to_def, feature_to_def
from feathr.registry._feature_registry_purview import _PurviewRegistry
from feathr.spark_provider._databricks_submission import _FeathrDatabricksJobLauncher
from feathr.spark_provider._localspark_submission import _FeathrLocalSparkJobLauncher
from feathr.spark_provider._synapse_submission import _FeathrSynapseJobLauncher
from feathr.spark_provider.feathr_configurations import SparkExecutionConfiguration
from feathr.udf._preprocessing_pyudf_manager import _PreprocessingPyudfManager
from feathr.utils._env_config_reader import EnvConfigReader
from feathr.utils._file_utils import write_to_file
from feathr.utils.feature_printer import FeaturePrinter
from feathr.utils.spark_job_params import FeatureGenerationJobParams, FeatureJoinJobParams
from feathr.version import get_version
import importlib.util



class FeathrClient(object):
    """Feathr client.

    The client is used to create training dataset, materialize features, register features, and fetch features from
    the online storage.

    For offline storage and compute engine, Azure ADLS, AWS S3 and Azure Synapse are supported.

    For online storage, currently only Redis is supported.
    The users of this client is responsible for set up all the necessary information needed to start a Redis client via
    environment variable or a Spark cluster. Host address, port and password are needed to start the Redis client.

    Raises:
        RuntimeError: Fail to create the client since necessary environment variables are not set for Redis
            client creation.
    """
    def __init__(
        self,
        config_path:str = "./feathr_config.yaml",
        local_workspace_dir: str = None,
        credential: Any = None,
        project_registry_tag: Dict[str, str] = None,
    ):
        """Initialize Feathr Client.
        Configuration values used by the Feathr are evaluated in the following precedence, with items higher on the list taking priority.
            1. Environment variables
            2. Values in the configuration file
            3. Values in the Azure Key Vault

        Args:
            config_path (optional): Config yaml file path. See [Feathr Config Template](https://github.com/feathr-ai/feathr/blob/main/feathr_project/feathrcli/data/feathr_user_workspace/feathr_config.yaml) for more details.  Defaults to "./feathr_config.yaml".
            local_workspace_dir (optional): Set where is the local work space dir. If not set, Feathr will create a temporary folder to store local workspace related files.
            credential (optional): Azure credential to access cloud resources, most likely to be the returned result of DefaultAzureCredential(). If not set, Feathr will initialize DefaultAzureCredential() inside the __init__ function to get credentials.
            project_registry_tag (optional): Adding tags for project in Feathr registry. This might be useful if you want to tag your project as deprecated, or allow certain customizations on project level. Default is empty
        """
        self.logger = logging.getLogger(__name__)
        # Redis key separator
        self._KEY_SEPARATOR = ':'
        self._COMPOSITE_KEY_SEPARATOR = '#'
        self.env_config = EnvConfigReader(config_path=config_path)
        if local_workspace_dir:
            self.local_workspace_dir = local_workspace_dir
        else:
            # this is required for Windows
            tem_dir_obj = tempfile.TemporaryDirectory()
            self.local_workspace_dir = tem_dir_obj.name

        if not os.path.exists(config_path):
            self.logger.warning('No Configuration file exist at the user provided config_path or the default config_path (./feathr_config.yaml), you need to set the environment variables explicitly. For all the environment variables that you need to set, please refer to https://github.com/feathr-ai/feathr/blob/main/feathr_project/feathrcli/data/feathr_user_workspace/feathr_config.yaml')

        # Load all configs from yaml at initialization
        # DO NOT load any configs from yaml during runtime.
        self.project_name = self.env_config.get(
            'project_config__project_name')

        # Redis configs. This is optional unless users have configured Redis host.
        if self.env_config.get('online_store__redis__host'):
            # For illustrative purposes.
            spec = importlib.util.find_spec("redis")
            if spec is None:
                self.logger.warning('You have configured Redis host, but there is no local Redis client package. Install the package using "pip install redis". ')
            self.redis_host = self.env_config.get(
                'online_store__redis__host')
            self.redis_port = self.env_config.get(
                'online_store__redis__port')
            self.redis_ssl_enabled = self.env_config.get(
                'online_store__redis__ssl_enabled')
            self._construct_redis_client()

        # Offline store enabled configs; false by default
        self.s3_enabled = self.env_config.get(
            'offline_store__s3__s3_enabled')
        self.adls_enabled = self.env_config.get(
            'offline_store__adls__adls_enabled')
        self.wasb_enabled = self.env_config.get(
            'offline_store__wasb__wasb_enabled')
        self.jdbc_enabled = self.env_config.get(
            'offline_store__jdbc__jdbc_enabled')
        self.snowflake_enabled = self.env_config.get(
            'offline_store__snowflake__snowflake_enabled')
        if not (self.s3_enabled or self.adls_enabled or self.wasb_enabled or self.jdbc_enabled or self.snowflake_enabled):
            self.logger.warning("No offline storage enabled.")

        # S3 configs
        if self.s3_enabled:
            self.s3_endpoint = self.env_config.get(
                'offline_store__s3__s3_endpoint')

        # spark configs
        self.output_num_parts = self.env_config.get(
            'spark_config__spark_result_output_parts')
        self.spark_runtime = self.env_config.get(
            'spark_config__spark_cluster')

        self.credential = credential
        if self.spark_runtime not in {'azure_synapse', 'databricks', 'local'}:
            raise RuntimeError(
                f'{self.spark_runtime} is not supported. Only \'azure_synapse\', \'databricks\' and \'local\' are currently supported.')
        elif self.spark_runtime == 'azure_synapse':
            # Feathr is a spark-based application so the feathr jar compiled from source code will be used in the
            # Spark job submission. The feathr jar hosted in cloud saves the time users needed to upload the jar from
            # their local env.
            self._FEATHR_JOB_JAR_PATH = \
                self.env_config.get(
                    'spark_config__azure_synapse__feathr_runtime_location')

            if self.credential is None:
                self.credential = DefaultAzureCredential(exclude_interactive_browser_credential=False)

            self.feathr_spark_launcher = _FeathrSynapseJobLauncher(
                synapse_dev_url=self.env_config.get(
                    'spark_config__azure_synapse__dev_url'),
                pool_name=self.env_config.get(
                    'spark_config__azure_synapse__pool_name'),
                datalake_dir=self.env_config.get(
                    'spark_config__azure_synapse__workspace_dir'),
                executor_size=self.env_config.get(
                    'spark_config__azure_synapse__executor_size'),
                executors=self.env_config.get(
                    'spark_config__azure_synapse__executor_num'),
                credential=self.credential
            )
        elif self.spark_runtime == 'databricks':
            # Feathr is a spark-based application so the feathr jar compiled from source code will be used in the
            # Spark job submission. The feathr jar hosted in cloud saves the time users needed to upload the jar from
            # their local env.
            self._FEATHR_JOB_JAR_PATH = \
                self.env_config.get(
                    'spark_config__databricks__feathr_runtime_location')

            self.feathr_spark_launcher = _FeathrDatabricksJobLauncher(
                workspace_instance_url=self.env_config.get(
                    'spark_config__databricks__workspace_instance_url'),
                token_value=self.env_config.get_from_env_or_akv(
                    'DATABRICKS_WORKSPACE_TOKEN_VALUE'),
                config_template=self.env_config.get(
                    'spark_config__databricks__config_template'),
                databricks_work_dir=self.env_config.get(
                    'spark_config__databricks__work_dir')
            )
        elif self.spark_runtime == 'local':
            self._FEATHR_JOB_JAR_PATH = \
                self.env_config.get(
                    'spark_config__local__feathr_runtime_location')
            self.feathr_spark_launcher = _FeathrLocalSparkJobLauncher(
                workspace_path = self.env_config.get('spark_config__local__workspace'),
                master = self.env_config.get('spark_config__local__master')
                )


        self.secret_names = []

        # initialize config helper
        self.config_helper = FeathrConfigHelper()

        # initialize registry
        self.registry = None
        registry_endpoint = self.env_config.get('feature_registry__api_endpoint')
        azure_purview_name = self.env_config.get('feature_registry__purview__purview_name')
        if registry_endpoint:
            self.registry = _FeatureRegistry(self.project_name, endpoint=registry_endpoint, project_tags=project_registry_tag, credential=credential)
        elif azure_purview_name:
            registry_delimiter = self.env_config.get('feature_registry__purview__delimiter')
            # initialize the registry no matter whether we set purview name or not, given some of the methods are used there.
            self.registry = _PurviewRegistry(self.project_name, azure_purview_name, registry_delimiter, project_registry_tag, config_path = config_path, credential=credential)
            logger.warning("FEATURE_REGISTRY__PURVIEW__PURVIEW_NAME will be deprecated soon. Please use FEATURE_REGISTRY__API_ENDPOINT instead.")
        else:
            # no registry configured
            logger.info("Feathr registry is not configured. Consider setting the Feathr registry component for richer feature store experience.")

        logger.info(f"Feathr client {get_version()} initialized successfully.")

    def _check_required_environment_variables_exist(self):
        """Checks if the required environment variables(form feathr_config.yaml) is set.

        Some required information has to be set via environment variables so the client can work.
        """
        props = self.secret_names
        for required_field in (self.required_fields + props):
            if required_field not in os.environ:
                raise RuntimeError(f'{required_field} is not set in environment variable. All required environment '
                                   f'variables are: {self.required_fields}.')

    def register_features(self, from_context: bool = True):
        """Registers features based on the current workspace

        Args:
            from_context: If from_context is True (default), the features will be generated from the current context, with the previous built features in client.build(). Otherwise, the features will be generated from
            configuration files.
        """

        if from_context:
            # make sure those items are in `self`
            if 'anchor_list' in dir(self) and 'derived_feature_list' in dir(self):
                self.config_helper.save_to_feature_config_from_context(self.anchor_list, self.derived_feature_list, self.local_workspace_dir)
                self.registry.register_features(self.local_workspace_dir, from_context=from_context, anchor_list=self.anchor_list, derived_feature_list=self.derived_feature_list)
            else:
                raise RuntimeError("Please call FeathrClient.build_features() first in order to register features")
        else:
            self.registry.register_features(self.local_workspace_dir, from_context=from_context)

    def build_features(self, anchor_list: List[FeatureAnchor] = [], derived_feature_list: List[DerivedFeature] = [], verbose: bool = False):
        """Build features based on the current workspace. all actions that triggers a spark job will be based on the
        result of this action.
        """
        # Run necessary validations
        # anchor name and source name should be unique
        anchor_names = {}
        source_names = {}
        for anchor in anchor_list:
            if anchor.name in anchor_names:
                raise RuntimeError(f"Anchor name should be unique but there are duplicate anchor names in your anchor "
                                   f"definitions. Anchor name of {anchor} is already defined in {anchor_names[anchor.name]}")
            else:
                anchor_names[anchor.name] = anchor
            if anchor.source.name in source_names and (anchor.source is not source_names[anchor.source.name]):
                raise RuntimeError(f"Source name should be unique but there are duplicate source names in your source "
                                   f"definitions. Source name of {anchor.source} is already defined in {source_names[anchor.source.name]}")
            else:
                source_names[anchor.source.name] = anchor.source

        _PreprocessingPyudfManager.build_anchor_preprocessing_metadata(anchor_list, self.local_workspace_dir)
        self.config_helper.save_to_feature_config_from_context(anchor_list, derived_feature_list, self.local_workspace_dir)
        self.anchor_list = anchor_list
        self.derived_feature_list = derived_feature_list

        # Check if data source used by every anchor requires additional system properties to be set
        props = []
        for anchor in self.anchor_list:
            if hasattr(anchor.source, "get_required_properties"):
                props.extend(anchor.source.get_required_properties())
        # Reset `system_properties`
        self.secret_names = props

        # Pretty print anchor_list
        if verbose and self.anchor_list:
            FeaturePrinter.pretty_print_anchors(self.anchor_list)

    def get_snowflake_path(self, database: str, schema: str, dbtable: str = None, query: str = None) -> str:
        """
        Returns snowflake path given dataset location information.
        Either dbtable or query must be specified but not both.
        """
        if dbtable is not None and query is not None:
            raise RuntimeError("Both dbtable and query are specified. Can only specify one..")
        if dbtable is None and query is None:
            raise RuntimeError("One of dbtable or query must be specified..")
        if dbtable:
            return f"snowflake://snowflake_account/?sfDatabase={database}&sfSchema={schema}&dbtable={dbtable}"
        else:
            return f"snowflake://snowflake_account/?sfDatabase={database}&sfSchema={schema}&query={query}"

    def list_registered_features(self, project_name: str = None) -> List[str]:
        """List all the already registered features under the given project.
        `project_name` must not be None or empty string because it violates the RBAC policy
        """
        return self.registry.list_registered_features(project_name)

    def list_dependent_entities(self, qualified_name: str):
        """
        Lists all dependent/downstream entities for a given entity
        """
        return self.registry.list_dependent_entities(qualified_name)

    def delete_entity(self, qualified_name: str):
        """
        Deletes a single entity if it has no downstream/dependent entities
        """
        return self.registry.delete_entity(qualified_name)

    def _get_registry_client(self):
        """
        Returns registry client in case users want to perform more advanced operations
        """
        return self.registry._get_registry_client()

    def get_online_features(self, feature_table: str, key: Any, feature_names: List[str]):
        """Fetches feature value for a certain key from a online feature table.

        Args:
            feature_table: the name of the feature table.
            key: the key/key list of the entity;
                 for key list, please make sure the order is consistent with the one in feature's definition;
                 the order can be found by 'get_features_from_registry'.
            feature_names: list of feature names to fetch

        Return:
            A list of feature values for this entity. It's ordered by the requested feature names.
            For example, feature_names = ['f_is_medium_trip_distance', 'f_day_of_week', 'f_day_of_month', 'f_hour_of_day']
            then, the returned feature values is: [b'true', b'4.0', b'31.0', b'23.0'].
            If the feature_table or key doesn't exist, then a list of Nones are returned. For example,
            [None, None, None, None].
            If a feature doesn't exist, then a None is returned for that feature. For example:
            [None, b'4.0', b'31.0', b'23.0'].
            """
        redis_key = self._construct_redis_key(feature_table, key)
        res = self.redis_client.hmget(redis_key, *feature_names)
        return self._decode_proto(res)

    def multi_get_online_features(self, feature_table: str, keys: List[Any], feature_names: List[str]):
        """Fetches feature value for a list of keys from a online feature table. This is the batch version of the get API.

        Args:
            feature_table: the name of the feature table.
            keys: list of keys/composite keys for the entities;
                  for composite keys, please make sure each order of them is consistent with the one in feature's definition;
                  the order can be found by 'get_features_from_registry'.
            feature_names: list of feature names to fetch

        Return:
            A list of feature values for the requested entities. It's ordered by the requested feature names. For
            example, keys = [12, 24], feature_names = ['f_is_medium_trip_distance', 'f_day_of_week', 'f_day_of_month',
            'f_hour_of_day'] then, the returned feature values is: {'12': [b'false', b'5.0', b'1.0', b'0.0'],
            '24': [b'true', b'4.0', b'31.0', b'23.0']}. If the feature_table or key doesn't exist, then a list of Nones
            are returned. For example, {'12': [None, None, None, None], '24': [None, None, None, None]} If a feature
            doesn't exist, then a None is returned for that feature. For example: {'12': [None, b'4.0', b'31.0',
            b'23.0'], '24': [b'true', b'4.0', b'31.0', b'23.0']}.
        """
        with self.redis_client.pipeline() as redis_pipeline:
            for key in keys:
                redis_key = self._construct_redis_key(feature_table, key)
                redis_pipeline.hmget(redis_key, *feature_names)
            pipeline_result = redis_pipeline.execute()

        decoded_pipeline_result = []
        for feature_list in pipeline_result:
            decoded_pipeline_result.append(self._decode_proto(feature_list))
        for i in range(len(keys)):
            if isinstance(keys[i], List):
                keys[i] = self._COMPOSITE_KEY_SEPARATOR.join(keys[i])
        return dict(zip(keys, decoded_pipeline_result))

    def _decode_proto(self, feature_list):
        """Decode the bytes(in string form) via base64 decoder. For dense array, it will be returned as Python List.
        For sparse array, it will be returned as tuple of index array and value array. The order of elements in the
        arrays won't be changed.
        """
        typed_result = []
        for raw_feature in feature_list:
            if raw_feature:
                feature_value = FeatureValue()
                decoded = base64.b64decode(raw_feature)
                feature_value.ParseFromString(decoded)
                if feature_value.WhichOneof('FeatureValueOneOf') == 'boolean_value':
                    typed_result.append(feature_value.boolean_value)
                elif feature_value.WhichOneof('FeatureValueOneOf') == 'string_value':
                    typed_result.append(feature_value.string_value)
                elif feature_value.WhichOneof('FeatureValueOneOf') == 'float_value':
                    typed_result.append(feature_value.float_value)
                elif feature_value.WhichOneof('FeatureValueOneOf') == 'double_value':
                    typed_result.append(feature_value.double_value)
                elif feature_value.WhichOneof('FeatureValueOneOf') == 'int_value':
                    typed_result.append(feature_value.int_value)
                elif feature_value.WhichOneof('FeatureValueOneOf') == 'long_value':
                    typed_result.append(feature_value.long_value)
                elif feature_value.WhichOneof('FeatureValueOneOf') == 'int_array':
                    typed_result.append(feature_value.int_array.integers)
                elif feature_value.WhichOneof('FeatureValueOneOf') == 'string_array':
                    typed_result.append(feature_value.string_array.strings)
                elif feature_value.WhichOneof('FeatureValueOneOf') == 'float_array':
                    typed_result.append(feature_value.float_array.floats)
                elif feature_value.WhichOneof('FeatureValueOneOf') == 'double_array':
                    typed_result.append(feature_value.double_array.doubles)
                elif feature_value.WhichOneof('FeatureValueOneOf') == 'boolean_array':
                    typed_result.append(feature_value.boolean_array.booleans)
                elif feature_value.WhichOneof('FeatureValueOneOf') == 'sparse_string_array':
                    typed_result.append((feature_value.sparse_string_array.index_integers, feature_value.sparse_string_array.value_strings))
                elif feature_value.WhichOneof('FeatureValueOneOf') == 'sparse_bool_array':
                    typed_result.append((feature_value.sparse_bool_array.index_integers, feature_value.sparse_bool_array.value_booleans))
                elif feature_value.WhichOneof('FeatureValueOneOf') == 'sparse_float_array':
                    typed_result.append((feature_value.sparse_float_array.index_integers, feature_value.sparse_float_array.value_floats))
                elif feature_value.WhichOneof('FeatureValueOneOf') == 'sparse_double_array':
                    typed_result.append((feature_value.sparse_double_array.index_integers, feature_value.sparse_double_array.value_doubles))
                elif feature_value.WhichOneof('FeatureValueOneOf') == 'sparse_long_array':
                    typed_result.append((feature_value.sparse_long_array.index_integers, feature_value.sparse_long_array.value_longs))
                else:
                    self.logger.debug("Fail to load the feature type. Maybe a new type that is not supported by this "
                                      "client version")
                    self.logger.debug(f"The raw feature is {raw_feature}.")
                    self.logger.debug(f"The loaded feature is {feature_value}")
                    typed_result.append(None)
            else:
                typed_result.append(raw_feature)
        return typed_result

    def delete_feature_from_redis(self, feature_table, key, feature_name) -> None:
        """
        Delete feature from Redis

        Args:
            feature_table: the name of the feature table
            key: the key of the entity
            feature_name: feature name to be deleted
        """

        redis_key = self._construct_redis_key(feature_table, key)
        if self.redis_client.hexists(redis_key, feature_name):
            self.redis_client.delete(redis_key, feature_name)
            print(f'Deletion successful. {feature_name} is deleted from Redis.')
        else:
            raise RuntimeError(f'Deletion failed. {feature_name} not found in Redis.')

    def _clean_test_data(self, feature_table):
        """
        WARNING: THIS IS ONLY USED FOR TESTING
        Clears a namespace in redis cache.
        This may be very time consuming.

        Args:
          feature_table: str, feature_table i.e your prefix before the separator in the Redis database.
        """
        cursor = '0'
        ns_keys = feature_table + '*'
        while cursor != 0:
            # 5000 count at a scan seems reasonable faster for our testing data
            cursor, keys = self.redis_client.scan(
                cursor=cursor, match=ns_keys, count=5000)
            if keys:
                self.redis_client.delete(*keys)

    def _construct_redis_key(self, feature_table, key):
        if isinstance(key, List):
            key = self._COMPOSITE_KEY_SEPARATOR.join(key)
        return feature_table + self._KEY_SEPARATOR + key

    def _str_to_bool(self, s: str, variable_name = None):
        """Define a function to detect convert string to bool, since Redis client sometimes require a bool and sometimes require a str
        """
        if (isinstance(s, str) and s.casefold() == 'True'.casefold()) or s == True:
            return True
        elif (isinstance(s, str) and s.casefold() == 'False'.casefold()) or s == False:
            return False
        else:
            self.logger.warning(f'{s} is not a valid Bool value. Maybe you want to double check if it is set correctly for {variable_name}.')
            return s

    def _construct_redis_client(self):
        """Constructs the Redis client. The host, port, credential and other parameters can be set via environment
        parameters.
        """
        password = self.env_config.get_from_env_or_akv(REDIS_PASSWORD)
        host = self.redis_host
        port = self.redis_port
        ssl_enabled = self.redis_ssl_enabled
        self.redis_client = redis.Redis(
            host=host,
            port=port,
            password=password,
            ssl=self._str_to_bool(ssl_enabled, "ssl_enabled"))
        self.logger.info('Redis connection is successful and completed.')

    def get_offline_features(self,
                             observation_settings: ObservationSettings,
                             feature_query: Union[FeatureQuery, List[FeatureQuery]],
                             output_path: Union[str, Sink],
                             execution_configurations: Union[SparkExecutionConfiguration ,Dict[str,str]] = {},
                             config_file_name:str = "feature_join_conf/feature_join.conf",
                             dataset_column_names: Set[str] = None,
                             verbose: bool = False
                             ):
        """
        Get offline features for the observation dataset
        Args:
            observation_settings: settings of the observation data, e.g. timestamp columns, input path, etc.
            feature_query: features that are requested to add onto the observation data
            output_path: output path of job, i.e. the observation data with features attached.
            execution_configurations: a dict that will be passed to spark job when the job starts up, i.e. the "spark configurations". Note that not all of the configuration will be honored since some of the configurations are managed by the Spark platform, such as Databricks or Azure Synapse. Refer to the [spark documentation](https://spark.apache.org/docs/latest/configuration.html) for a complete list of spark configurations.
            config_file_name: the name of the config file that will be passed to the spark job. The config file is used to configure the spark job. The default value is "feature_join_conf/feature_join.conf".
            dataset_column_names: column names of observation data set. Will be used to check conflicts with feature names if cannot get real column names from observation data set.
        """
        feature_queries = feature_query if isinstance(feature_query, List) else [feature_query]
        feature_names = []
        for feature_query in feature_queries:
            for feature_name in feature_query.feature_list:
                feature_names.append(feature_name)
        
        if len(feature_names) > 0 and observation_settings.conflicts_auto_correction is None:
            import feathr.utils.job_utils as job_utils
            dataset_column_names_from_path = job_utils.get_cloud_file_column_names(self, observation_settings.observation_path, observation_settings.file_format,observation_settings.is_file_path)
            if (dataset_column_names_from_path is None or len(dataset_column_names_from_path) == 0) and dataset_column_names is None:
                self.logger.warning(f"Feathr is unable to read the Observation data from {observation_settings.observation_path} due to permission issue or invalid path. Please either grant the permission or supply the observation column names in the filed: observation_column_names.")
            else:
                if dataset_column_names_from_path is not None and len(dataset_column_names_from_path) > 0:
                    dataset_column_names = dataset_column_names_from_path
                conflict_names = []
                for feature_name in feature_names:
                    if feature_name in dataset_column_names:
                        conflict_names.append(feature_name)
                if len(conflict_names) != 0:
                    conflict_names = ",".join(conflict_names)
                    raise RuntimeError(f"Feature names exist conflicts with dataset column names: {conflict_names}")
           
        udf_files = _PreprocessingPyudfManager.prepare_pyspark_udf_files(feature_names, self.local_workspace_dir)

        # produce join config
        tm = Template("""
            {{observation_settings.to_feature_config()}}
            featureList: [
                {% for list in feature_lists %}
                    {{list.to_feature_config()}}
                {% endfor %}
            ]
            outputPath: "{{output_path}}"
        """)
        config = tm.render(feature_lists=feature_queries, observation_settings=observation_settings, output_path=output_path)
        config_file_path = os.path.join(self.local_workspace_dir, config_file_name)

        # make sure `FeathrClient.build_features()` is called before getting offline features/materialize features
        # otherwise users will be confused on what are the available features
        # in build_features it will assign anchor_list and derived_feature_list variable, hence we are checking if those two variables exist to make sure the above condition is met
        if 'anchor_list' in dir(self) and 'derived_feature_list' in dir(self):
            self.config_helper.save_to_feature_config_from_context(self.anchor_list, self.derived_feature_list, self.local_workspace_dir)
        else:
            raise RuntimeError("Please call FeathrClient.build_features() first in order to get offline features")

        # Pretty print feature_query
        if verbose and feature_query:
            FeaturePrinter.pretty_print_feature_query(feature_query)

        write_to_file(content=config, full_file_name=config_file_path)
        return self._get_offline_features_with_config(config_file_path,
                                                      output_path=output_path,
                                                      execution_configurations=execution_configurations,
                                                      udf_files=udf_files)

    def _get_offline_features_with_config(self,
                                          feature_join_conf_path='feature_join_conf/feature_join.conf',
                                          output_path: Union[str, Sink] = "",
                                          execution_configurations: Dict[str,str] = {},
                                          udf_files=[]):
        """Joins the features to your offline observation dataset based on the join config.

        Args:
          feature_join_conf_path: Relative path to your feature join config file.
        """
        cloud_udf_paths = [self.feathr_spark_launcher.upload_or_get_cloud_path(udf_local_path) for udf_local_path in udf_files]
        feathr_feature = ConfigFactory.parse_file(feature_join_conf_path)

        feature_join_job_params = FeatureJoinJobParams(join_config_path=os.path.abspath(feature_join_conf_path),
                                                       observation_path=feathr_feature['observationPath'],
                                                       feature_config=os.path.join(self.local_workspace_dir, 'feature_conf/'),
                                                       job_output_path=output_path)
        job_tags = { OUTPUT_PATH_TAG: feature_join_job_params.job_output_path }
        # set output format in job tags if it's set by user, so that it can be used to parse the job result in the helper function
        if execution_configurations is not None and OUTPUT_FORMAT in execution_configurations:
            job_tags[OUTPUT_FORMAT] = execution_configurations[OUTPUT_FORMAT]
        else:
            job_tags[OUTPUT_FORMAT] = "avro"
        '''
        - Job tags are for job metadata and it's not passed to the actual spark job (i.e. not visible to spark job), more like a platform related thing that Feathr want to add (currently job tags only have job output URL and job output format, ). They are carried over with the job and is visible to every Feathr client. Think this more like some customized metadata for the job which would be weird to be put in the spark job itself.
        - Job arguments (or sometimes called job parameters)are the arguments which are command line arguments passed into the actual spark job. This is usually highly related with the spark job. In Feathr it's like the input to the scala spark CLI. They are usually not spark specific (for example if we want to specify the location of the feature files, or want to
        - Job configuration are like "configurations" for the spark job and are usually spark specific. For example, we want to control the no. of write parts for spark
        Job configurations and job arguments (or sometimes called job parameters) have quite some overlaps (i.e. you can achieve the same goal by either using the job arguments/parameters vs. job configurations). But the job tags should just be used for metadata purpose.
        '''

        # submit the jars
        return self.feathr_spark_launcher.submit_feathr_job(
            job_name=self.project_name + '_feathr_feature_join_job',
            main_jar_path=self._FEATHR_JOB_JAR_PATH,
            python_files=cloud_udf_paths,
            job_tags=job_tags,
            main_class_name=JOIN_CLASS_NAME,
            arguments= [
                '--join-config', self.feathr_spark_launcher.upload_or_get_cloud_path(
                    feature_join_job_params.join_config_path),
                '--input', feature_join_job_params.observation_path,
                '--output', feature_join_job_params.job_output_path,
                '--feature-config', self.feathr_spark_launcher.upload_or_get_cloud_path(
                    feature_join_job_params.feature_config),
                '--num-parts', self.output_num_parts
            ]+self._get_offline_storage_arguments(),
            reference_files_path=[],
            configuration=execution_configurations,
            properties=self._collect_secrets(feature_join_job_params.secrets)
        )

    def _get_offline_storage_arguments(self):
        arguments = []
        if self.s3_enabled:
            arguments.append('--s3-config')
            arguments.append(self._get_s3_config_str())
        if self.adls_enabled:
            arguments.append('--adls-config')
            arguments.append(self._get_adls_config_str())
        if self.wasb_enabled:
            arguments.append('--blob-config')
            arguments.append(self._get_blob_config_str())
        if self.jdbc_enabled:
            arguments.append('--sql-config')
            arguments.append(self._get_sql_config_str())
        if self.snowflake_enabled:
            arguments.append('--snowflake-config')
            arguments.append(self._get_snowflake_config_str())
        return arguments

    def get_job_result_uri(self, block=True, timeout_sec=300) -> str:
        """Gets the job output URI
        """
        if not block:
            return self.feathr_spark_launcher.get_job_result_uri()
        # Block the API by pooling the job status and wait for complete
        if self.feathr_spark_launcher.wait_for_completion(timeout_sec):
            return self.feathr_spark_launcher.get_job_result_uri()
        else:
            raise RuntimeError(
                'Spark job failed so output cannot be retrieved.')

    def get_job_tags(self) -> Dict[str, str]:
        """Gets the job tags
        """
        return self.feathr_spark_launcher.get_job_tags()

    def wait_job_to_finish(self, timeout_sec: int = 300):
        """Waits for the job to finish in a blocking way unless it times out
        """
        if self.feathr_spark_launcher.wait_for_completion(timeout_sec):
            return
        else:
            raise RuntimeError('Spark job failed.')

    def monitor_features(self, settings: MonitoringSettings, execution_configurations: Union[SparkExecutionConfiguration ,Dict[str,str]] = {}, verbose: bool = False):
        """Create a offline job to generate statistics to monitor feature data

        Args:
            settings: Feature monitoring settings
            execution_configurations: a dict that will be passed to spark job when the job starts up, i.e. the "spark configurations". Note that not all of the configuration will be honored since some of the configurations are managed by the Spark platform, such as Databricks or Azure Synapse. Refer to the [spark documentation](https://spark.apache.org/docs/latest/configuration.html) for a complete list of spark configurations.
        """
        self.materialize_features(settings, execution_configurations, verbose)

    # Get feature keys given the name of a feature
    # Should search in both 'derived_feature_list' and 'anchor_list'
    # Return related keys(key_column list) or None if cannot find the feature
    def _get_feature_key(self, feature_name: str):
        features = []
        if 'derived_feature_list' in dir(self):
            features += self.derived_feature_list
        if 'anchor_list' in dir(self):
            for anchor in self.anchor_list:
                features += anchor.features
        for feature in features:
            if feature.name == feature_name:
                keys = feature.key
                return set(key.key_column for key in keys)
        self.logger.warning(f"Invalid feature name: {feature_name}. Please call FeathrClient.build_features() first in order to materialize the features.")
        return None

    # Validation on feature keys:
    # Features within a set of aggregation or planned to be merged should have same keys
    # The param "allow_empty_key" shows if empty keys are acceptable
    def _valid_materialize_keys(self, features: List[str], allow_empty_key=False):
        keys = None
        for feature in features:
            new_keys = self._get_feature_key(feature)
            if new_keys is None:
                self.logger.error(f"Key of feature: {feature} is empty. Please confirm the feature is defined. In addition, if this feature is not from INPUT_CONTEXT, you might want to double check on the feature definition to see whether the key is empty or not.")
                return False
            # If only get one key and it's "NOT_NEEDED", it means the feature has an empty key.
            if ','.join(new_keys) == "NOT_NEEDED" and not allow_empty_key:
                self.logger.error(f"Empty feature key is not allowed for features: {features}")
                return False
            if keys is None:
                keys = copy.deepcopy(new_keys)
            else:
                if len(keys) != len(new_keys):
                    self.logger.error(f"Inconsistent feature keys. Current keys are {str(keys)}")
                    return False
                for new_key in new_keys:
                    if new_key not in keys:
                        self.logger.error(f"Inconsistent feature keys. Current keys are {str(keys)}")
                        return False
        return True

    def materialize_features(self, settings: MaterializationSettings, execution_configurations: Union[SparkExecutionConfiguration ,Dict[str,str]] = {}, verbose: bool = False, allow_materialize_non_agg_feature: bool = False):
        """Materialize feature data

        Args:
            settings: Feature materialization settings
            execution_configurations: a dict that will be passed to spark job when the job starts up, i.e. the "spark configurations". Note that not all of the configuration will be honored since some of the configurations are managed by the Spark platform, such as Databricks or Azure Synapse. Refer to the [spark documentation](https://spark.apache.org/docs/latest/configuration.html) for a complete list of spark configurations.
            allow_materialize_non_agg_feature: Materializing non-aggregated features (the features without WindowAggTransformation) doesn't output meaningful results so it's by default set to False, but if you really want to materialize non-aggregated features, set this to True.
        """
        feature_list = settings.feature_names
        if len(feature_list) > 0:
            if 'anchor_list' in dir(self):
                anchors = [anchor for anchor in self.anchor_list if isinstance(anchor.source, InputContext)]
                anchor_feature_names = set(feature.name for anchor in anchors for feature in anchor.features)
                for feature in feature_list:
                    if feature in anchor_feature_names:
                        raise RuntimeError(f"Materializing features that are defined on INPUT_CONTEXT is not supported. {feature} is defined on INPUT_CONTEXT so you should remove it from the feature list in MaterializationSettings.")
            if not self._valid_materialize_keys(feature_list):
                raise RuntimeError(f"Invalid materialization features: {feature_list}, since they have different keys or they are not defined. Currently Feathr only supports materializing features of the same keys.")

        if not allow_materialize_non_agg_feature:
            # Check if there are non-aggregation features in the list
            for fn in feature_list:
                # Check over anchor features
                for anchor in self.anchor_list:
                    for feature in anchor.features:
                        if feature.name == fn and not isinstance(feature.transform, WindowAggTransformation):
                            raise RuntimeError(f"Feature {fn} is not an aggregation feature. Currently Feathr only supports materializing aggregation features. If you want to materialize {fn}, please set allow_materialize_non_agg_feature to True.")
                # Check over derived features
                for feature in self.derived_feature_list:
                    if feature.name == fn and not isinstance(feature.transform, WindowAggTransformation):
                        raise RuntimeError(f"Feature {fn} is not an aggregation feature. Currently Feathr only supports materializing aggregation features. If you want to materialize {fn}, please set allow_materialize_non_agg_feature to True.")

        # Collect secrets from sinks. Get output_path as well if the sink is offline sink (HdfsSink) for later use.
        secrets = []
        output_path = None
        for sink in settings.sinks:
            if hasattr(sink, "get_required_properties"):
                secrets.extend(sink.get_required_properties())
            if isinstance(sink, HdfsSink):
                # Note, for now we only cache one output path from one of HdfsSinks (if one passed multiple sinks).
                output_path = sink.output_path

        results = []
        # produce materialization config
        for end in settings.get_backfill_cutoff_time():
            settings.backfill_time.end = end
            config = _to_materialization_config(settings)
            config_file_name = "feature_gen_conf/auto_gen_config_{}.conf".format(end.timestamp())
            config_file_path = os.path.join(self.local_workspace_dir, config_file_name)
            write_to_file(content=config, full_file_name=config_file_path)

            # make sure `FeathrClient.build_features()` is called before getting offline features/materialize features in the python SDK
            # otherwise users will be confused on what are the available features
            # in build_features it will assign anchor_list and derived_feature_list variable, hence we are checking if those two variables exist to make sure the above condition is met
            if 'anchor_list' in dir(self) and 'derived_feature_list' in dir(self):
                self.config_helper.save_to_feature_config_from_context(self.anchor_list, self.derived_feature_list, self.local_workspace_dir)
            else:
                raise RuntimeError("Please call FeathrClient.build_features() first in order to materialize the features")

            udf_files = _PreprocessingPyudfManager.prepare_pyspark_udf_files(settings.feature_names, self.local_workspace_dir)
            # CLI will directly call this so the experience won't be broken
            result = self._materialize_features_with_config(
                feature_gen_conf_path=config_file_path,
                execution_configurations=execution_configurations,
                udf_files=udf_files,
                secrets=secrets,
                output_path=output_path,
            )
            if os.path.exists(config_file_path) and self.spark_runtime != 'local':
                os.remove(config_file_path)
            results.append(result)

        # Pretty print feature_names of materialized features
        if verbose and settings:
            FeaturePrinter.pretty_print_materialize_features(settings)

        return results

    def _materialize_features_with_config(
        self,
        feature_gen_conf_path: str = 'feature_gen_conf/feature_gen.conf',
        execution_configurations: Dict[str,str] = {},
        udf_files: List = [],
        secrets: List = [],
        output_path: str = None,
    ):
        """Materializes feature data based on the feature generation config. The feature
        data will be materialized to the destination specified in the feature generation config.

        Args
            feature_gen_conf_path: Relative path to the feature generation config you want to materialize.
            execution_configurations: Spark job execution configurations.
            udf_files: UDF files.
            secrets: Secrets to access sinks.
            output_path: The output path of the materialized features when using an offline sink.
        """
        cloud_udf_paths = [self.feathr_spark_launcher.upload_or_get_cloud_path(udf_local_path) for udf_local_path in udf_files]

        # Read all features conf
        generation_config = FeatureGenerationJobParams(
            generation_config_path=os.path.abspath(feature_gen_conf_path),
            feature_config=os.path.join(self.local_workspace_dir, "feature_conf/"))

        # When using offline sink (i.e. output_path is not None)
        job_tags = {}
        if output_path:
            job_tags[OUTPUT_PATH_TAG] = output_path
            # set output format in job tags if it's set by user, so that it can be used to parse the job result in the helper function
            if execution_configurations is not None and OUTPUT_FORMAT in execution_configurations:
                job_tags[OUTPUT_FORMAT] = execution_configurations[OUTPUT_FORMAT]
            else:
                job_tags[OUTPUT_FORMAT] = "avro"
        '''
        - Job tags are for job metadata and it's not passed to the actual spark job (i.e. not visible to spark job), more like a platform related thing that Feathr want to add (currently job tags only have job output URL and job output format, ). They are carried over with the job and is visible to every Feathr client. Think this more like some customized metadata for the job which would be weird to be put in the spark job itself.
        - Job arguments (or sometimes called job parameters)are the arguments which are command line arguments passed into the actual spark job. This is usually highly related with the spark job. In Feathr it's like the input to the scala spark CLI. They are usually not spark specific (for example if we want to specify the location of the feature files, or want to
        - Job configuration are like "configurations" for the spark job and are usually spark specific. For example, we want to control the no. of write parts for spark
        Job configurations and job arguments (or sometimes called job parameters) have quite some overlaps (i.e. you can achieve the same goal by either using the job arguments/parameters vs. job configurations). But the job tags should just be used for metadata purpose.
        '''

        optional_params = []
        if self.env_config.get_from_env_or_akv('KAFKA_SASL_JAAS_CONFIG'):
            optional_params = optional_params + ['--kafka-config', self._get_kafka_config_str()]
        arguments = [
                '--generation-config', self.feathr_spark_launcher.upload_or_get_cloud_path(
                generation_config.generation_config_path),
                # Local Config, comma seperated file names
                '--feature-config', self.feathr_spark_launcher.upload_or_get_cloud_path(
                generation_config.feature_config),
                '--redis-config', self._getRedisConfigStr(),
            ] + self._get_offline_storage_arguments()+optional_params
        monitoring_config_str = self._get_monitoring_config_str()
        if monitoring_config_str:
            arguments.append('--monitoring-config')
            arguments.append(monitoring_config_str)
        return self.feathr_spark_launcher.submit_feathr_job(
            job_name=self.project_name + '_feathr_feature_materialization_job',
            main_jar_path=self._FEATHR_JOB_JAR_PATH,
            python_files=cloud_udf_paths,
            job_tags=job_tags,
            main_class_name=GEN_CLASS_NAME,
            arguments=arguments,
            reference_files_path=[],
            configuration=execution_configurations,
            properties=self._collect_secrets(secrets)
        )

    def wait_job_to_finish(self, timeout_sec: int = 300):
        """Waits for the job to finish in a blocking way unless it times out
        """
        if self.feathr_spark_launcher.wait_for_completion(timeout_sec):
            return
        else:
            raise RuntimeError('Spark job failed.')

    def _getRedisConfigStr(self):
        """Construct the Redis config string. The host, port, credential and other parameters can be set via environment
        variables."""
        password = self.env_config.get_from_env_or_akv(REDIS_PASSWORD)
        host = self.redis_host
        port = self.redis_port
        ssl_enabled = self.redis_ssl_enabled
        config_str = """
        REDIS_PASSWORD: "{REDIS_PASSWORD}"
        REDIS_HOST: "{REDIS_HOST}"
        REDIS_PORT: {REDIS_PORT}
        REDIS_SSL_ENABLED: {REDIS_SSL_ENABLED}
        """.format(REDIS_PASSWORD=password, REDIS_HOST=host, REDIS_PORT=port, REDIS_SSL_ENABLED=str(ssl_enabled))
        return self._reshape_config_str(config_str)

    def _get_s3_config_str(self):
        """Construct the S3 config string. The endpoint, access key, secret key, and other parameters can be set via
        environment variables."""
        endpoint = self.s3_endpoint
        # if s3 endpoint is set in the feathr_config, then we need other environment variables
        # keys can't be only accessed through environment
        access_key = self.env_config.get_from_env_or_akv('S3_ACCESS_KEY')
        secret_key = self.env_config.get_from_env_or_akv('S3_SECRET_KEY')
        # HOCON format will be parsed by the Feathr job
        config_str = """
            S3_ENDPOINT: {S3_ENDPOINT}
            S3_ACCESS_KEY: "{S3_ACCESS_KEY}"
            S3_SECRET_KEY: "{S3_SECRET_KEY}"
            """.format(S3_ENDPOINT=endpoint, S3_ACCESS_KEY=access_key, S3_SECRET_KEY=secret_key)
        return self._reshape_config_str(config_str)

    def _get_adls_config_str(self):
        """Construct the ADLS config string for abfs(s). The Account, access key and other parameters can be set via
        environment variables."""
        account = self.env_config.get_from_env_or_akv('ADLS_ACCOUNT')
        # if ADLS Account is set in the feathr_config, then we need other environment variables
        # keys can't be only accessed through environment
        key = self.env_config.get_from_env_or_akv('ADLS_KEY')
        # HOCON format will be parsed by the Feathr job
        config_str = """
            ADLS_ACCOUNT: {ADLS_ACCOUNT}
            ADLS_KEY: "{ADLS_KEY}"
            """.format(ADLS_ACCOUNT=account, ADLS_KEY=key)
        return self._reshape_config_str(config_str)

    def _get_blob_config_str(self):
        """Construct the Blob config string for wasb(s). The Account, access key and other parameters can be set via
        environment variables."""
        account = self.env_config.get_from_env_or_akv('BLOB_ACCOUNT')
        # if BLOB Account is set in the feathr_config, then we need other environment variables
        # keys can't be only accessed through environment
        key = self.env_config.get_from_env_or_akv('BLOB_KEY')
        # HOCON format will be parsed by the Feathr job
        config_str = """
            BLOB_ACCOUNT: {BLOB_ACCOUNT}
            BLOB_KEY: "{BLOB_KEY}"
            """.format(BLOB_ACCOUNT=account, BLOB_KEY=key)
        return self._reshape_config_str(config_str)

    def _get_sql_config_str(self):
        """Construct the SQL config string for jdbc. The dbtable (query), user, password and other parameters can be set via
        environment variables."""
        table = self.env_config.get_from_env_or_akv('JDBC_TABLE')
        user = self.env_config.get_from_env_or_akv('JDBC_USER')
        password = self.env_config.get_from_env_or_akv('JDBC_PASSWORD')
        driver = self.env_config.get_from_env_or_akv('JDBC_DRIVER')
        auth_flag = self.env_config.get_from_env_or_akv('JDBC_AUTH_FLAG')
        token = self.env_config.get_from_env_or_akv('JDBC_TOKEN')
        # HOCON format will be parsed by the Feathr job
        config_str = """
            JDBC_TABLE: {JDBC_TABLE}
            JDBC_USER: {JDBC_USER}
            JDBC_PASSWORD: {JDBC_PASSWORD}
            JDBC_DRIVER: {JDBC_DRIVER}
            JDBC_AUTH_FLAG: {JDBC_AUTH_FLAG}
            JDBC_TOKEN: {JDBC_TOKEN}
            """.format(JDBC_TABLE=table, JDBC_USER=user, JDBC_PASSWORD=password, JDBC_DRIVER = driver, JDBC_AUTH_FLAG = auth_flag, JDBC_TOKEN = token)
        return self._reshape_config_str(config_str)

    def _get_monitoring_config_str(self):
        """Construct monitoring-related config string."""
        url = self.env_config.get('monitoring__database__sql__url')
        user = self.env_config.get('monitoring__database__sql__user')
        password = self.env_config.get_from_env_or_akv('MONITORING_DATABASE_SQL_PASSWORD')
        if url:
            # HOCON format will be parsed by the Feathr job
            config_str = """
                MONITORING_DATABASE_SQL_URL: "{url}"
                MONITORING_DATABASE_SQL_USER: {user}
                MONITORING_DATABASE_SQL_PASSWORD: {password}
                """.format(url=url, user=user, password=password)
            return self._reshape_config_str(config_str)
        else:
            ""

    def _get_snowflake_config_str(self):
        """Construct the Snowflake config string for jdbc. The url, user, role and other parameters can be set via
        yaml config. Password can be set via environment variables."""
        sf_url = self.env_config.get('offline_store__snowflake__url')
        sf_user = self.env_config.get('offline_store__snowflake__user')
        sf_role = self.env_config.get('offline_store__snowflake__role')
        sf_warehouse = self.env_config.get('offline_store__snowflake__warehouse')
        sf_password = self.env_config.get_from_env_or_akv('JDBC_SF_PASSWORD')
        # HOCON format will be parsed by the Feathr job
        config_str = """
            JDBC_SF_URL: {JDBC_SF_URL}
            JDBC_SF_USER: {JDBC_SF_USER}
            JDBC_SF_ROLE: {JDBC_SF_ROLE}
            JDBC_SF_WAREHOUSE: {JDBC_SF_WAREHOUSE}
            JDBC_SF_PASSWORD: {JDBC_SF_PASSWORD}
            """.format(JDBC_SF_URL=sf_url, JDBC_SF_USER=sf_user, JDBC_SF_PASSWORD=sf_password, JDBC_SF_ROLE=sf_role, JDBC_SF_WAREHOUSE=sf_warehouse)
        return self._reshape_config_str(config_str)

    def _get_kafka_config_str(self):
        """Construct the Kafka config string. The endpoint, access key, secret key, and other parameters can be set via
        environment variables."""
        sasl = self.env_config.get_from_env_or_akv('KAFKA_SASL_JAAS_CONFIG')
        # HOCON format will be parsed by the Feathr job
        config_str = """
            KAFKA_SASL_JAAS_CONFIG: "{sasl}"
            """.format(sasl=sasl)
        return self._reshape_config_str(config_str)

    def _collect_secrets(self, additional_secrets=[]):
        """Collect all values corresponding to the secret names."""
        prop_and_value = {}
        for prop in self.secret_names + additional_secrets:
            prop = prop.upper()
            prop_and_value[prop] = self.env_config.get(prop)
        return prop_and_value

    def get_features_from_registry(self, project_name: str, return_keys: bool = False, verbose: bool = False) -> Union[Dict[str, FeatureBase], Tuple[Dict[str, FeatureBase], Dict[str, Union[TypedKey, List[TypedKey]]]]]:
        """
        Get feature from registry by project name. The features got from registry are automatically built.
        """
        registry_anchor_list, registry_derived_feature_list = self.registry.get_features_from_registry(project_name)
        self.build_features(registry_anchor_list, registry_derived_feature_list)
        feature_dict = {}
        key_dict = {}
        # add those features into a dict for easier lookup
        if verbose and registry_anchor_list:
            logger.info("Get anchor features from registry: ")
        for anchor in registry_anchor_list:
            for feature in anchor.features:
                feature_dict[feature.name] = feature
                key_dict[feature.name] = feature.key
                if verbose:
                    logger.info(json.dumps(feature_to_def(feature), indent=2))
        if verbose and registry_derived_feature_list:
            logger.info("Get derived features from registry: ")
        for feature in registry_derived_feature_list:
                feature_dict[feature.name] = feature
                key_dict[feature.name] = feature.key
                if verbose:
                    logger.info(json.dumps(derived_feature_to_def(feature), indent=2))
        if return_keys:
            return feature_dict, key_dict
        return feature_dict

    def _reshape_config_str(self, config_str:str):
        if self.spark_runtime == 'local':
            return "'{" + config_str + "}'"
        else:
            return config_str
 


from feathr import ObservationSettings


def test_observation_setting_with_timestamp():
    observation_settings = ObservationSettings(
        observation_path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04.csv",
        event_timestamp_column="lpep_dropoff_datetime",
        timestamp_format="yyyy-MM-dd HH:mm:ss")
    config = observation_settings.to_feature_config()
    expected_config = """
            settings: {
                joinTimeSettings: {
                    timestampColumn: {
                        def: "lpep_dropoff_datetime"
                        format: "yyyy-MM-dd HH:mm:ss"
                    }
                }
            }
            
            observationPath: "wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04.csv"
        """
    assert ''.join(config.split()) == ''.join(expected_config.split())


def test_observation_setting_without_timestamp():

    observation_settings = ObservationSettings(
        observation_path='snowflake://snowflake_account/?dbtable=CALL_CENTER&sfDatabase=SNOWFLAKE_SAMPLE_DATA&sfSchema=TPCDS_SF10TCL')
    config = observation_settings.to_feature_config()
    expected_config = """            
            observationPath:"snowflake://snowflake_account/?dbtable=CALL_CENTER&sfDatabase=SNOWFLAKE_SAMPLE_DATA&sfSchema=TPCDS_SF10TCL"
        """
    assert ''.join(config.split()) == ''.join(expected_config.split())


from datetime import datetime, timedelta
from typing import List, Optional
from feathr.definition.sink import HdfsSink, RedisSink, Sink
import math


class BackfillTime:
    """Time range to materialize/backfill feature data. Please refer to https://feathr-ai.github.io/feathr/concepts/materializing-features.html#feature-backfill for a more detailed explanation.

    Attributes:
        start: start time of the backfill, inclusive.
        end: end time of the backfill, inclusive.
        step: duration of each backfill step. e.g. if you want to materialize features on daily basis, use timedelta(days=1)
    """
    def __init__(self, start: datetime, end: datetime, step: timedelta):
        self.start = start
        self.end = end
        self.step = step


class MaterializationSettings:
    """Settings about materialization features.

    Attributes:
        name: The materialization job name
        sinks: sinks where the materialized features should be written to
        feature_names: list of feature names to be materialized
        backfill_time: time range and frequency for the materialization. Default to now().
        resolution: time interval for output directories. Only support 'DAILY' and 'HOURLY' for now (DAILY by default).
                    If 'DAILY', output paths should be: yyyy/MM/dd;
                    Otherwise would be: yyyy/MM/dd/HH
    """
    def __init__(self, name: str, sinks: List[Sink], feature_names: List[str], backfill_time: Optional[BackfillTime] = None, resolution: str = "DAILY"):
        if resolution not in ["DAILY", "HOURLY"]:
            raise RuntimeError(
                f'{resolution} is not supported. Only \'DAILY\' and \'HOURLY\' are currently supported.')
        self.resolution = resolution
        self.name = name
        now = datetime.now()
        self.backfill_time = backfill_time if backfill_time else BackfillTime(start=now, end=now, step=timedelta(days=1))
        for sink in sinks:
            if isinstance(sink, HdfsSink):
                self.has_hdfs_sink = True
                sink.aggregation_features = feature_names
            elif isinstance(sink, RedisSink):
                sink.aggregation_features = feature_names
        self.sinks = sinks
        self.feature_names = feature_names

    def get_backfill_cutoff_time(self) -> List[datetime]:
        """Get the backfill cutoff time points for materialization. 
        E.g. for `BackfillTime(start=datetime(2022, 3, 1), end=datetime(2022, 3, 5), step=timedelta(days=1))`, 
        it returns cutoff time list as `[2022-3-1, 2022-3-2, 2022-3-3, 2022-3-4, 2022-3-5]`, 
        for `BackfillTime(start=datetime(2022, 3, 1, 1), end=datetime(2022, 3, 1, 5), step=timedelta(hours=1))`, 
        it returns cutoff time list as `[2022-3-1 01:00:00, 2022-3-1 02:00:00, 2022-3-1 03:00:00, 2022-3-1 04:00:00, 2022-3-1 05:00:00]`
        """        

        start_time = self.backfill_time.start
        end_time = self.backfill_time.end
        step_in_seconds = self.backfill_time.step.total_seconds()
        assert start_time <= end_time, "Start time {} must be earlier or equal to end time {}".format(start_time, end_time)
        assert step_in_seconds > 0, "Step in time range should be greater than 0, but got {}".format(step_in_seconds)
        num_delta = (self.backfill_time.end - self.backfill_time.start).total_seconds() / step_in_seconds
        num_delta = math.floor(num_delta) + 1
        return [end_time - timedelta(seconds=n*step_in_seconds) for n in reversed(range(num_delta))]


"""Utilities for downloading sample datasets"""

from feathr.datasets.constants import (
    NYC_TAXI_SMALL_URL,
    FRAUD_DETECTION_ACCOUNT_INFO_URL,
    FRAUD_DETECTION_FRAUD_TRANSACTIONS_URL,
    FRAUD_DETECTION_UNTAGGED_TRANSACTIONS_URL,
    PRODUCT_RECOMMENDATION_USER_OBSERVATION_URL,
    PRODUCT_RECOMMENDATION_USER_PROFILE_URL,
    PRODUCT_RECOMMENDATION_USER_PURCHASE_HISTORY_URL,
    PRODUCT_RECOMMENDATION_PRODUCT_DETAIL_URL,
    HOTEL_REVIEWS_URL,
)

__all__ = [
    "NYC_TAXI_SMALL_URL",
    "FRAUD_DETECTION_ACCOUNT_INFO_URL",
    "FRAUD_DETECTION_FRAUD_TRANSACTIONS_URL",
    "FRAUD_DETECTION_UNTAGGED_TRANSACTIONS_URL",
    "PRODUCT_RECOMMENDATION_USER_OBSERVATION_URL",
    "PRODUCT_RECOMMENDATION_USER_PROFILE_URL",
    "PRODUCT_RECOMMENDATION_USER_PURCHASE_HISTORY_URL",
    "PRODUCT_RECOMMENDATION_PRODUCT_DETAIL_URL",
    "HOTEL_REVIEWS_URL",
]
